"""Tests for pipette state accessors in the protocol_engine state store.

DEPRECATED: Testing PipetteView independently of PipetteStore is no longer helpful.
Try to add new tests to test_pipette_state.py, where they can be tested together,
treating PipetteState as a private implementation detail.
"""

from collections import OrderedDict
from typing import Dict, List, NamedTuple, Optional, Tuple, cast

import pytest
from decoy import Decoy

from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.pipette.pipette_definition import (
    AvailableSensorDefinition,
    ValidNozzleMaps,
)
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

from ..pipette_fixtures import (
    EIGHT_CHANNEL_COLS,
    EIGHT_CHANNEL_MAP,
    EIGHT_CHANNEL_ROWS,
    NINETY_SIX_COLS,
    NINETY_SIX_MAP,
    NINETY_SIX_ROWS,
    get_default_nozzle_map,
)
from opentrons.config.defaults_ot2 import Z_RETRACT_DISTANCE
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine import errors
from opentrons.protocol_engine.errors import PipetteNotLoadedError, TipNotAttachedError
from opentrons.protocol_engine.state import fluid_stack
from opentrons.protocol_engine.state.pipettes import (
    BoundingNozzlesOffsets,
    CurrentDeckPoint,
    HardwarePipette,
    PipetteBoundingBoxOffsets,
    PipetteState,
    PipetteView,
    StaticPipetteConfig,
)
from opentrons.protocol_engine.types import (
    CurrentPipetteLocation,
    DeckPoint,
    FlowRates,
    LabwareWellId,
    LoadedPipette,
    MotorAxis,
    TipGeometry,
)
from opentrons.types import Mount as HwMount
from opentrons.types import MountType, NozzleConfigurationType, Point

_SAMPLE_NOZZLE_BOUNDS_OFFSETS = BoundingNozzlesOffsets(
    back_left_offset=Point(x=10, y=20, z=30), front_right_offset=Point(x=40, y=50, z=60)
)
_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS = PipetteBoundingBoxOffsets(
    back_left_corner=Point(x=10, y=20, z=30),
    front_right_corner=Point(x=40, y=50, z=60),
    front_left_corner=Point(x=10, y=50, z=60),
    back_right_corner=Point(x=40, y=20, z=60),
)


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


def get_pipette_view(
    pipettes_by_id: Optional[Dict[str, LoadedPipette]] = None,
    current_well: Optional[CurrentPipetteLocation] = None,
    current_deck_point: CurrentDeckPoint = CurrentDeckPoint(
        mount=None, deck_point=None
    ),
    attached_tip_by_id: Optional[Dict[str, Optional[TipGeometry]]] = None,
    movement_speed_by_id: Optional[Dict[str, Optional[float]]] = None,
    static_config_by_id: Optional[Dict[str, StaticPipetteConfig]] = None,
    flow_rates_by_id: Optional[Dict[str, FlowRates]] = None,
    nozzle_layout_by_id: Optional[Dict[str, NozzleMap]] = None,
    liquid_presence_detection_by_id: Optional[Dict[str, bool]] = None,
    ready_to_aspirate_by_id: Optional[Dict[str, bool]] = None,
    pipette_contents_by_id: Optional[
        Dict[str, Optional[fluid_stack.FluidStack]]
    ] = None,
    has_clean_tips_by_id: Optional[Dict[str, bool]] = None,
    last_tip_rack_well_by_id: Optional[Dict[str, Optional[LabwareWellId]]] = None,
) -> PipetteView:
    """Get a pipette view test subject with the specified state."""
    state = PipetteState(
        pipettes_by_id=pipettes_by_id or {},
        pipette_contents_by_id=pipette_contents_by_id or {},
        current_location=current_well,
        current_deck_point=current_deck_point,
        attached_tip_by_id=attached_tip_by_id or {},
        movement_speed_by_id=movement_speed_by_id or {},
        static_config_by_id=static_config_by_id or {},
        flow_rates_by_id=flow_rates_by_id or {},
        nozzle_configuration_by_id=nozzle_layout_by_id or {},
        liquid_presence_detection_by_id=liquid_presence_detection_by_id or {},
        ready_to_aspirate_by_id=ready_to_aspirate_by_id or {},
        has_clean_tips_by_id=has_clean_tips_by_id or {},
        tip_source_by_id=last_tip_rack_well_by_id or {},
    )

    return PipetteView(state=state)


def create_pipette_config(
    name: str,
    back_compat_names: Optional[List[str]] = None,
    ready_to_aspirate: bool = False,
) -> PipetteDict:
    """Create a fake but valid (enough) PipetteDict object."""
    return cast(
        PipetteDict,
        {
            "name": name,
            "back_compat_names": back_compat_names or [],
            "ready_to_aspirate": ready_to_aspirate,
        },
    )


def test_initial_pipette_data_by_id() -> None:
    """It should should raise if pipette ID doesn't exist."""
    subject = get_pipette_view()

    with pytest.raises(errors.PipetteNotLoadedError):
        subject.get("asdfghjkl")


def test_initial_pipette_data_by_mount() -> None:
    """It should return None if mount isn't present."""
    subject = get_pipette_view()

    assert subject.get_by_mount(MountType.LEFT) is None
    assert subject.get_by_mount(MountType.RIGHT) is None


def test_get_pipette_data() -> None:
    """It should get pipette data by ID and mount from the state."""
    pipette_data = LoadedPipette(
        id="pipette-id",
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    subject = get_pipette_view(pipettes_by_id={"pipette-id": pipette_data})

    result_by_id = subject.get("pipette-id")
    result_by_mount = subject.get_by_mount(MountType.LEFT)

    assert result_by_id == pipette_data
    assert result_by_mount == pipette_data
    assert subject.get_mount("pipette-id") == MountType.LEFT


def test_get_hardware_pipette() -> None:
    """It maps a pipette ID to a config given the HC's attached pipettes."""
    pipette_config = create_pipette_config("p300_single")
    attached_pipettes: Dict[HwMount, PipetteDict] = {
        HwMount.LEFT: pipette_config,
        HwMount.RIGHT: cast(PipetteDict, {}),
    }

    subject = get_pipette_view(
        pipettes_by_id={
            "left-id": LoadedPipette(
                id="left-id",
                mount=MountType.LEFT,
                pipetteName=PipetteNameType.P300_SINGLE,
            ),
            "right-id": LoadedPipette(
                id="right-id",
                mount=MountType.RIGHT,
                pipetteName=PipetteNameType.P300_MULTI,
            ),
        }
    )

    result = subject.get_hardware_pipette(
        pipette_id="left-id",
        attached_pipettes=attached_pipettes,
    )

    assert result == HardwarePipette(mount=HwMount.LEFT, config=pipette_config)

    with pytest.raises(errors.PipetteNotAttachedError):
        subject.get_hardware_pipette(
            pipette_id="right-id",
            attached_pipettes=attached_pipettes,
        )


def test_get_hardware_pipette_with_back_compat() -> None:
    """It maps a pipette ID to a config given the HC's attached pipettes.

    In this test, the hardware config specs "p300_single_gen2", and the
    loaded pipette name in state is "p300_single," which is is allowed.
    """
    pipette_config = create_pipette_config(
        "p300_single_gen2",
        back_compat_names=["p300_single"],
    )
    attached_pipettes: Dict[HwMount, PipetteDict] = {
        HwMount.LEFT: pipette_config,
        HwMount.RIGHT: cast(PipetteDict, {}),
    }

    subject = get_pipette_view(
        pipettes_by_id={
            "pipette-id": LoadedPipette(
                id="pipette-id",
                mount=MountType.LEFT,
                pipetteName=PipetteNameType.P300_SINGLE,
            ),
        }
    )

    result = subject.get_hardware_pipette(
        pipette_id="pipette-id",
        attached_pipettes=attached_pipettes,
    )

    assert result == HardwarePipette(mount=HwMount.LEFT, config=pipette_config)


def test_get_hardware_pipette_raises_with_name_mismatch() -> None:
    """It maps a pipette ID to a config given the HC's attached pipettes.

    In this test, the hardware config specs "p300_single_gen2", but the
    loaded pipette name in state is "p10_single," which does not match.
    """
    pipette_config = create_pipette_config("p300_single_gen2")
    attached_pipettes: Dict[HwMount, Optional[PipetteDict]] = {
        HwMount.LEFT: pipette_config,
        HwMount.RIGHT: cast(PipetteDict, {}),
    }

    subject = get_pipette_view(
        pipettes_by_id={
            "pipette-id": LoadedPipette(
                id="pipette-id",
                mount=MountType.LEFT,
                pipetteName=PipetteNameType.P10_SINGLE,
            ),
        }
    )

    with pytest.raises(errors.PipetteNotAttachedError):
        subject.get_hardware_pipette(
            pipette_id="pipette-id",
            attached_pipettes=attached_pipettes,
        )


def test_get_aspirated_volume(decoy: Decoy) -> None:
    """It should get the aspirate volume for a pipette."""
    stack = decoy.mock(cls=fluid_stack.FluidStack)
    subject = get_pipette_view(
        pipette_contents_by_id={
            "pipette-id": stack,
            "pipette-id-none": None,
            "pipette-id-no-tip": None,
        },
        attached_tip_by_id={
            "pipette-id": TipGeometry(length=1, volume=2, diameter=3),
            "pipette-id-none": TipGeometry(length=4, volume=5, diameter=6),
            "pipette-id-no-tip": None,
        },
    )
    decoy.when(stack.aspirated_volume()).then_return(42)

    assert subject.get_aspirated_volume("pipette-id") == 42
    assert subject.get_aspirated_volume("pipette-id-none") is None

    with pytest.raises(errors.PipetteNotLoadedError):
        subject.get_aspirated_volume("not-an-id")

    with pytest.raises(errors.TipNotAttachedError):
        subject.get_aspirated_volume("pipette-id-no-tip")


def test_get_pipette_working_volume(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should get the minimum value of tip volume and max volume."""
    subject = get_pipette_view(
        attached_tip_by_id={
            "pipette-id": TipGeometry(length=1, volume=1337, diameter=42.0),
        },
        static_config_by_id={
            "pipette-id": StaticPipetteConfig(
                min_volume=1,
                max_volume=9001,
                channels=5,
                model="blah",
                display_name="bleh",
                serial_number="",
                tip_configuration_lookup_table={9001: supported_tip_fixture},
                nominal_tip_overlap={},
                home_position=0,
                nozzle_offset_z=0,
                bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
                default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
                pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
                lld_settings={},
                plunger_positions={
                    "top": 0.0,
                    "bottom": 5.0,
                    "blow_out": 19.0,
                    "drop_tip": 20.0,
                },
                shaft_ul_per_mm=5.0,
                available_sensors=available_sensors,
                volume_mode=VolumeModes.default,
                available_volume_modes_min_vol={},
            )
        },
    )

    assert subject.get_working_volume("pipette-id") == 1337


def test_get_pipette_working_volume_raises_if_tip_volume_is_none(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """Should raise an exception that no tip is attached."""
    subject = get_pipette_view(
        attached_tip_by_id={
            "pipette-id": None,
        },
        static_config_by_id={
            "pipette-id": StaticPipetteConfig(
                min_volume=1,
                max_volume=9001,
                channels=5,
                model="blah",
                display_name="bleh",
                serial_number="",
                tip_configuration_lookup_table={9001: supported_tip_fixture},
                nominal_tip_overlap={},
                home_position=0,
                nozzle_offset_z=0,
                bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
                default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
                pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
                lld_settings={},
                plunger_positions={
                    "top": 0.0,
                    "bottom": 5.0,
                    "blow_out": 19.0,
                    "drop_tip": 20.0,
                },
                shaft_ul_per_mm=5.0,
                available_sensors=available_sensors,
                volume_mode=VolumeModes.default,
                available_volume_modes_min_vol={},
            )
        },
    )

    with pytest.raises(TipNotAttachedError):
        subject.get_working_volume("pipette-id")

    with pytest.raises(PipetteNotLoadedError):
        subject.get_working_volume("wrong-id")


def test_get_pipette_available_volume(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    decoy: Decoy,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should get the available volume for a pipette."""
    stack = decoy.mock(cls=fluid_stack.FluidStack)
    decoy.when(stack.aspirated_volume()).then_return(58)
    subject = get_pipette_view(
        attached_tip_by_id={
            "pipette-id": TipGeometry(
                length=1,
                diameter=2,
                volume=100,
            ),
        },
        pipette_contents_by_id={"pipette-id": stack},
        static_config_by_id={
            "pipette-id": StaticPipetteConfig(
                min_volume=1,
                max_volume=123,
                channels=3,
                model="blah",
                display_name="bleh",
                serial_number="",
                tip_configuration_lookup_table={123: supported_tip_fixture},
                nominal_tip_overlap={},
                home_position=0,
                nozzle_offset_z=0,
                bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
                default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
                pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
                lld_settings={},
                plunger_positions={
                    "top": 0.0,
                    "bottom": 5.0,
                    "blow_out": 19.0,
                    "drop_tip": 20.0,
                },
                shaft_ul_per_mm=5.0,
                available_sensors=available_sensors,
                volume_mode=VolumeModes.default,
                available_volume_modes_min_vol={},
            ),
            "pipette-id-none": StaticPipetteConfig(
                min_volume=1,
                max_volume=123,
                channels=5,
                model="blah",
                display_name="bleh",
                serial_number="",
                tip_configuration_lookup_table={123: supported_tip_fixture},
                nominal_tip_overlap={},
                home_position=0,
                nozzle_offset_z=0,
                bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
                default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
                pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
                lld_settings={},
                plunger_positions={
                    "top": 0.0,
                    "bottom": 5.0,
                    "blow_out": 19.0,
                    "drop_tip": 20.0,
                },
                shaft_ul_per_mm=5.0,
                available_sensors=available_sensors,
                volume_mode=VolumeModes.default,
                available_volume_modes_min_vol={},
            ),
        },
    )

    assert subject.get_available_volume("pipette-id") == 42


def test_get_attached_tip() -> None:
    """It should get the tip-rack ID map of a pipette's attached tip."""
    subject = get_pipette_view(
        attached_tip_by_id={
            "foo": TipGeometry(length=1, volume=2, diameter=3),
            "bar": None,
        }
    )

    assert subject.get_attached_tip("foo") == TipGeometry(
        length=1, volume=2, diameter=3
    )
    assert subject.get_attached_tip("bar") is None
    assert subject.get_all_attached_tips() == [
        ("foo", TipGeometry(length=1, volume=2, diameter=3)),
    ]


def test_validate_tip_state() -> None:
    """It should validate a pipette's tip attached state."""
    subject = get_pipette_view(
        attached_tip_by_id={
            "has-tip": TipGeometry(length=1, volume=2, diameter=3),
            "no-tip": None,
        }
    )

    subject.validate_tip_state(pipette_id="has-tip", expected_has_tip=True)
    subject.validate_tip_state(pipette_id="no-tip", expected_has_tip=False)

    with pytest.raises(errors.TipAttachedError):
        subject.validate_tip_state(pipette_id="has-tip", expected_has_tip=False)

    with pytest.raises(errors.TipNotAttachedError):
        subject.validate_tip_state(pipette_id="no-tip", expected_has_tip=True)


def test_get_movement_speed() -> None:
    """It should return the movement speed that was set for the given pipette."""
    subject = get_pipette_view(
        movement_speed_by_id={
            "pipette-with-movement-speed": 123.456,
            "pipette-without-movement-speed": None,
        }
    )

    assert (
        subject.get_movement_speed(pipette_id="pipette-with-movement-speed") == 123.456
    )
    assert (
        subject.get_movement_speed(pipette_id="pipette-without-movement-speed") is None
    )


@pytest.mark.parametrize(
    ("mount", "deck_point", "expected_deck_point"),
    [
        (MountType.LEFT, DeckPoint(x=1, y=2, z=3), DeckPoint(x=1, y=2, z=3)),
        (MountType.LEFT, None, None),
        (MountType.RIGHT, DeckPoint(x=1, y=2, z=3), None),
        (None, DeckPoint(x=1, y=2, z=3), None),
        (None, None, None),
    ],
)
def test_get_deck_point(
    mount: Optional[MountType],
    deck_point: Optional[DeckPoint],
    expected_deck_point: Optional[DeckPoint],
) -> None:
    """It should return the deck point for the given pipette."""
    pipette_data = LoadedPipette(
        id="pipette-id",
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    subject = get_pipette_view(
        pipettes_by_id={"pipette-id": pipette_data},
        current_deck_point=CurrentDeckPoint(
            mount=MountType.LEFT, deck_point=DeckPoint(x=1, y=2, z=3)
        ),
    )

    assert subject.get_deck_point(pipette_id="pipette-id") == DeckPoint(x=1, y=2, z=3)


def test_get_static_config(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return the static pipette configuration that was set for the given pipette."""
    config = StaticPipetteConfig(
        model="pipette-model",
        display_name="display name",
        serial_number="serial-number",
        min_volume=1.23,
        max_volume=4.56,
        channels=9,
        tip_configuration_lookup_table={4.56: supported_tip_fixture},
        nominal_tip_overlap={},
        home_position=10.12,
        nozzle_offset_z=12.13,
        bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
        default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
        lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={},
    )

    subject = get_pipette_view(
        pipettes_by_id={
            "pipette-id": LoadedPipette(
                id="pipette-id",
                mount=MountType.LEFT,
                pipetteName=PipetteNameType.P300_SINGLE,
            )
        },
        attached_tip_by_id={
            "pipette-id": TipGeometry(length=1, volume=4.56, diameter=3),
        },
        static_config_by_id={"pipette-id": config},
    )

    assert subject.get_config("pipette-id") == config
    assert subject.get_model_name("pipette-id") == "pipette-model"
    assert subject.get_display_name("pipette-id") == "display name"
    assert subject.get_serial_number("pipette-id") == "serial-number"
    assert subject.get_minimum_volume("pipette-id") == 1.23
    assert subject.get_maximum_volume("pipette-id") == 4.56
    assert subject.get_return_tip_scale("pipette-id") == 0.5
    assert (
        subject.get_instrument_max_height_ot2("pipette-id")
        == 22.25 - Z_RETRACT_DISTANCE
    )


def test_get_nominal_tip_overlap(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should return the static pipette configuration that was set for the given pipette."""
    config = StaticPipetteConfig(
        model="",
        display_name="",
        serial_number="",
        min_volume=0,
        max_volume=0,
        channels=10,
        tip_configuration_lookup_table={0: supported_tip_fixture},
        nominal_tip_overlap={
            "some-uri": 100,
            "default": 10,
        },
        home_position=0,
        nozzle_offset_z=0,
        bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
        default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        pipette_bounding_box_offsets=_SAMPLE_PIPETTE_BOUNDING_BOX_OFFSETS,
        lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={},
    )

    subject = get_pipette_view(static_config_by_id={"pipette-id": config})

    assert subject.get_nominal_tip_overlap("pipette-id", "some-uri") == 100
    assert subject.get_nominal_tip_overlap("pipette-id", "missing-uri") == 10


@pytest.mark.parametrize(
    ("mount", "expected_z_axis", "expected_plunger_axis"),
    [
        (MountType.LEFT, MotorAxis.LEFT_Z, MotorAxis.LEFT_PLUNGER),
        (MountType.RIGHT, MotorAxis.RIGHT_Z, MotorAxis.RIGHT_PLUNGER),
    ],
)
def test_get_motor_axes(
    mount: MountType, expected_z_axis: MotorAxis, expected_plunger_axis: MotorAxis
) -> None:
    """It should get a pipette's motor axes."""
    subject = get_pipette_view(
        pipettes_by_id={
            "pipette-id": LoadedPipette(
                id="pipette-id",
                mount=mount,
                pipetteName=PipetteNameType.P300_SINGLE,
            ),
        },
    )

    assert subject.get_z_axis("pipette-id") == expected_z_axis
    assert subject.get_plunger_axis("pipette-id") == expected_plunger_axis


def test_nozzle_configuration_getters() -> None:
    """Test that pipette view returns correct nozzle configuration data."""
    nozzle_map = NozzleMap.build(
        physical_nozzles=OrderedDict({"A1": Point(0, 0, 0)}),
        physical_rows=OrderedDict({"A": ["A1"]}),
        physical_columns=OrderedDict({"1": ["A1"]}),
        starting_nozzle="A1",
        back_left_nozzle="A1",
        front_right_nozzle="A1",
        valid_nozzle_maps=ValidNozzleMaps(maps={"A1": ["A1"]}),
    )
    subject = get_pipette_view(nozzle_layout_by_id={"pipette-id": nozzle_map})
    assert subject.get_nozzle_configuration("pipette-id") == nozzle_map
    assert subject.get_nozzle_configurations() == {"pipette-id": nozzle_map}
    assert subject.get_nozzle_layout_type("pipette-id") == NozzleConfigurationType.FULL
    assert subject.get_is_partially_configured("pipette-id") is False
    assert subject.get_primary_nozzle("pipette-id") == "A1"


class _PipetteSpecs(NamedTuple):
    tip_length: float
    bounding_box_offsets: PipetteBoundingBoxOffsets
    nozzle_map: NozzleMap
    critical_point: Optional[CriticalPoint]
    destination_position: Point
    pipette_bounds_result: Tuple[Point, Point, Point, Point]


_pipette_spec_cases = [
    _PipetteSpecs(
        # 8-channel P300, full configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(0.0, 31.5, 35.52),
            front_right_corner=Point(0.0, -31.5, 35.52),
            front_left_corner=Point(0.0, -31.5, 35.52),
            back_right_corner=Point(0.0, 31.5, 35.52),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=EIGHT_CHANNEL_MAP,
            physical_rows=EIGHT_CHANNEL_ROWS,
            physical_columns=EIGHT_CHANNEL_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Full": EIGHT_CHANNEL_COLS["1"]}),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            (
                Point(x=100.0, y=200.0, z=342.0),
                Point(x=100.0, y=137.0, z=342.0),
                Point(x=100.0, y=200.0, z=342.0),
                Point(x=100.0, y=137.0, z=342.0),
            )
        ),
    ),
    _PipetteSpecs(
        # 8-channel P300, single configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(0.0, 31.5, 35.52),
            front_right_corner=Point(0.0, -31.5, 35.52),
            front_left_corner=Point(0.0, -31.5, 35.52),
            back_right_corner=Point(0.0, 31.5, 35.52),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=EIGHT_CHANNEL_MAP,
            physical_rows=EIGHT_CHANNEL_ROWS,
            physical_columns=EIGHT_CHANNEL_COLS,
            starting_nozzle="H1",
            back_left_nozzle="H1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(maps={"H1": ["H1"]}),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            (
                Point(x=100.0, y=263.0, z=342.0),
                Point(x=100.0, y=200.0, z=342.0),
                Point(x=100.0, y=263.0, z=342.0),
                Point(x=100.0, y=200.0, z=342.0),
            )
        ),
    ),
    _PipetteSpecs(
        # 8-channel P300, full configuration. Critical point of XY_CENTER
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(0.0, 31.5, 35.52),
            front_right_corner=Point(0.0, -31.5, 35.52),
            front_left_corner=Point(0.0, -31.5, 35.52),
            back_right_corner=Point(0.0, 31.5, 35.52),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=EIGHT_CHANNEL_MAP,
            physical_rows=EIGHT_CHANNEL_ROWS,
            physical_columns=EIGHT_CHANNEL_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Full": EIGHT_CHANNEL_COLS["1"]}),
        ),
        critical_point=CriticalPoint.XY_CENTER,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            (
                Point(x=100.0, y=231.5, z=342.0),
                Point(x=100.0, y=168.5, z=342.0),
                Point(x=100.0, y=231.5, z=342.0),
                Point(x=100.0, y=168.5, z=342.0),
            )
        ),
    ),
    _PipetteSpecs(
        # 8-channel P300, Partial A1-E1 configuration. Critical point of XY_CENTER
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(0.0, 31.5, 35.52),
            front_right_corner=Point(0.0, -31.5, 35.52),
            front_left_corner=Point(0.0, -31.5, 35.52),
            back_right_corner=Point(0.0, 31.5, 35.52),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=EIGHT_CHANNEL_MAP,
            physical_rows=EIGHT_CHANNEL_ROWS,
            physical_columns=EIGHT_CHANNEL_COLS,
            starting_nozzle="H1",
            back_left_nozzle="E1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(
                maps={
                    "H1toE1": ["E1", "F1", "G1", "H1"],
                }
            ),
        ),
        critical_point=CriticalPoint.XY_CENTER,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            (
                Point(x=100.0, y=249.5, z=342.0),
                Point(x=100.0, y=186.5, z=342.0),
                Point(x=100.0, y=249.5, z=342.0),
                Point(x=100.0, y=186.5, z=342.0),
            )
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, full configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="H12",
            valid_nozzle_maps=ValidNozzleMaps(
                maps={
                    "Full": sum(
                        [
                            NINETY_SIX_ROWS["A"],
                            NINETY_SIX_ROWS["B"],
                            NINETY_SIX_ROWS["C"],
                            NINETY_SIX_ROWS["D"],
                            NINETY_SIX_ROWS["E"],
                            NINETY_SIX_ROWS["F"],
                            NINETY_SIX_ROWS["G"],
                            NINETY_SIX_ROWS["H"],
                        ],
                        [],
                    )
                }
            ),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            (
                Point(x=100.0, y=200.0, z=342.0),
                Point(x=199.0, y=137.0, z=342.0),
                Point(x=199.0, y=200.0, z=342.0),
                Point(x=100.0, y=137.0, z=342.0),
            )
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, A1 COLUMN configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Column1": NINETY_SIX_COLS["1"]}),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(100, 200, 342),
            Point(199, 137, 342),
            Point(199, 200, 342),
            Point(100, 137, 342),
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, A12 COLUMN configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A12",
            back_left_nozzle="A12",
            front_right_nozzle="H12",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Column12": NINETY_SIX_COLS["12"]}),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(1, 200, 342),
            Point(100, 137, 342),
            Point(100, 200, 342),
            Point(1, 137, 342),
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, ROW configuration
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="A12",
            valid_nozzle_maps=ValidNozzleMaps(maps={"RowA": NINETY_SIX_ROWS["A"]}),
        ),
        critical_point=None,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(100, 200, 342),
            Point(199, 137, 342),
            Point(199, 200, 342),
            Point(100, 137, 342),
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, ROW configuration. Critical point of XY_CENTER
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="A12",
            valid_nozzle_maps=ValidNozzleMaps(maps={"RowA": NINETY_SIX_ROWS["A"]}),
        ),
        critical_point=CriticalPoint.XY_CENTER,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(x=50.5, y=200, z=342),
            Point(x=149.5, y=137, z=342),
            Point(x=149.5, y=200, z=342),
            Point(x=50.5, y=137, z=342),
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, A12 COLUMN configuration. Critical point of Y_CENTER
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A12",
            back_left_nozzle="A12",
            front_right_nozzle="H12",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Column12": NINETY_SIX_COLS["12"]}),
        ),
        critical_point=CriticalPoint.Y_CENTER,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(1, 231.5, 342),
            Point(100, 168.5, 342),
            Point(100, 231.5, 342),
            Point(1, 168.5, 342),
        ),
    ),
    _PipetteSpecs(
        # 96-channel P1000, A1 COLUMN configuration. Critical point of XY_CENTER
        tip_length=42,
        bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(-36.0, -25.5, -259.15),
            front_right_corner=Point(63.0, -88.5, -259.15),
            front_left_corner=Point(-36.0, -88.5, -259.15),
            back_right_corner=Point(63.0, -25.5, -259.15),
        ),
        nozzle_map=NozzleMap.build(
            physical_nozzles=NINETY_SIX_MAP,
            physical_rows=NINETY_SIX_ROWS,
            physical_columns=NINETY_SIX_COLS,
            starting_nozzle="A1",
            back_left_nozzle="A1",
            front_right_nozzle="H1",
            valid_nozzle_maps=ValidNozzleMaps(maps={"Column1": NINETY_SIX_COLS["1"]}),
        ),
        critical_point=CriticalPoint.XY_CENTER,
        destination_position=Point(100, 200, 300),
        pipette_bounds_result=(
            Point(100, 231.5, 342),
            Point(199, 168.5, 342),
            Point(199, 231.5, 342),
            Point(100, 168.5, 342),
        ),
    ),
]


@pytest.mark.parametrize(
    argnames=_PipetteSpecs._fields,
    argvalues=_pipette_spec_cases,
)
def test_get_pipette_bounds_at_location(
    tip_length: float,
    bounding_box_offsets: PipetteBoundingBoxOffsets,
    nozzle_map: NozzleMap,
    destination_position: Point,
    critical_point: Optional[CriticalPoint],
    pipette_bounds_result: Tuple[Point, Point, Point, Point],
    available_sensors: AvailableSensorDefinition,
) -> None:
    """It should get the pipette's nozzle's bounds at the given location."""
    subject = get_pipette_view(
        nozzle_layout_by_id={"pipette-id": nozzle_map},
        attached_tip_by_id={
            "pipette-id": TipGeometry(length=tip_length, diameter=123, volume=123),
        },
        static_config_by_id={
            "pipette-id": StaticPipetteConfig(
                min_volume=1,
                max_volume=9001,
                channels=5,
                model="blah",
                display_name="bleh",
                serial_number="",
                tip_configuration_lookup_table={},
                nominal_tip_overlap={},
                home_position=0,
                nozzle_offset_z=0,
                default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
                bounding_nozzle_offsets=_SAMPLE_NOZZLE_BOUNDS_OFFSETS,
                pipette_bounding_box_offsets=bounding_box_offsets,
                lld_settings={},
                plunger_positions={
                    "top": 0.0,
                    "bottom": 5.0,
                    "blow_out": 19.0,
                    "drop_tip": 20.0,
                },
                shaft_ul_per_mm=5.0,
                available_sensors=available_sensors,
                volume_mode=VolumeModes.default,
                available_volume_modes_min_vol={},
            )
        },
    )
    assert (
        subject.get_pipette_bounds_at_specified_move_to_position(
            pipette_id="pipette-id",
            destination_position=destination_position,
            critical_point=critical_point,
        )
        == pipette_bounds_result
    )


@pytest.mark.parametrize(
    "nozzle_map,allowed",
    [
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H12",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={
                        "Full": sum(
                            [
                                NINETY_SIX_ROWS["A"],
                                NINETY_SIX_ROWS["B"],
                                NINETY_SIX_ROWS["C"],
                                NINETY_SIX_ROWS["D"],
                                NINETY_SIX_ROWS["E"],
                                NINETY_SIX_ROWS["F"],
                                NINETY_SIX_ROWS["G"],
                                NINETY_SIX_ROWS["H"],
                            ],
                            [],
                        )
                    }
                ),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H1",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"Column1": NINETY_SIX_COLS["1"]}
                ),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A12",
                back_left_nozzle="A12",
                front_right_nozzle="H12",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"Column12": NINETY_SIX_COLS["12"]}
                ),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="A1",
                valid_nozzle_maps=ValidNozzleMaps(maps={"Single": ["A1"]}),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=OrderedDict((("A1", Point(0.0, 1.0, 2.0)),)),
                physical_rows=OrderedDict((("1", ["A1"]),)),
                physical_columns=OrderedDict((("A", ["A1"]),)),
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="A1",
                valid_nozzle_maps=ValidNozzleMaps(maps={"Single": ["A1"]}),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=EIGHT_CHANNEL_MAP,
                physical_rows=EIGHT_CHANNEL_ROWS,
                physical_columns=EIGHT_CHANNEL_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H1",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"Full": EIGHT_CHANNEL_COLS["1"]}
                ),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=EIGHT_CHANNEL_MAP,
                physical_rows=EIGHT_CHANNEL_ROWS,
                physical_columns=EIGHT_CHANNEL_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="A1",
                valid_nozzle_maps=ValidNozzleMaps(maps={"Full": ["A1"]}),
            ),
            True,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A5",
                back_left_nozzle="A5",
                front_right_nozzle="H5",
                valid_nozzle_maps=ValidNozzleMaps(
                    maps={"Column12": NINETY_SIX_COLS["5"]}
                ),
            ),
            False,
        ),
    ],
)
def test_lld_config_validation(nozzle_map: NozzleMap, allowed: bool) -> None:
    """It should validate partial tip configurations for LLD."""
    pipette_id = "pipette-id"
    subject = get_pipette_view(
        nozzle_layout_by_id={pipette_id: nozzle_map},
    )
    assert subject.get_nozzle_configuration_supports_lld(pipette_id) == allowed



"""Tests for the JSON JsonTranslator interface."""

from typing import Dict, List

import pytest

from opentrons_shared_data.labware.labware_definition import (
    BrandData,
    CircularWellDefinition2,
    Dimensions,
    DisplayCategory,
    Group,
    GroupMetadata,
    LabwareDefinition,
    LabwareDefinition2,
    Metadata,
    Parameters2,
)
from opentrons_shared_data.labware.labware_definition import (
    Vector3D as SD_Labware_Vector,
)
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.protocol.models import (
    Labware,
    Liquid,
    Location,
    Module,
    OffsetVector,
    Pipette,
    ProfileStep,
    Robot,
    protocol_schema_v6,
    protocol_schema_v7,
    protocol_schema_v8,
)
from opentrons_shared_data.protocol.models import (
    Metadata as SD_Metadata,
)
from opentrons_shared_data.protocol.models import (
    WellLocation as SD_WellLocation,
)

from opentrons.protocol_engine import (
    DeckPoint,
    DeckSlotLocation,
    DropTipWellLocation,
    DropTipWellOrigin,
    LiquidHandlingWellLocation,
    ModuleLocation,
    ModuleModel,
    PickUpTipWellLocation,
    WellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine import (
    Liquid as PE_Liquid,
)
from opentrons.protocol_engine import (
    commands as pe_commands,
)
from opentrons.protocol_engine.types import HexColor
from opentrons.protocol_runner.json_translator import JsonTranslator
from opentrons.types import DeckSlotName, MountType

VALID_TEST_PARAMS = [
    (
        protocol_schema_v6.Command(
            commandType="aspirate",
            key=None,
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
            ),
        ),
        protocol_schema_v7.Command(
            commandType="aspirate",
            key=None,
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
            ),
        ),
        protocol_schema_v8.Command(
            commandType="aspirate",
            key=None,
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "volume": 1.23,
                "flowRate": 4.56,
                "wellName": "A1",
                "wellLocation": {
                    "origin": "bottom",
                    "offset": {"x": 0, "y": 0, "z": 7.89},
                },
            },
        ),
        pe_commands.AspirateCreate(
            key=None,
            params=pe_commands.AspirateParams(
                # todo: id
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=LiquidHandlingWellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=WellOffset(x=0, y=0, z=7.89),
                ),
            ),
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="dispense",
            key="dispense-key",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
            ),
        ),
        protocol_schema_v7.Command(
            commandType="dispense",
            key="dispense-key",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
            ),
        ),
        protocol_schema_v8.Command(
            commandType="dispense",
            key="dispense-key",
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "volume": 1.23,
                "flowRate": 4.56,
                "wellName": "A1",
                "wellLocation": {
                    "origin": "bottom",
                    "offset": {"x": 0, "y": 0, "z": 7.89},
                },
            },
        ),
        pe_commands.DispenseCreate(
            key="dispense-key",
            params=pe_commands.DispenseParams(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                volume=1.23,
                flowRate=4.56,
                wellName="A1",
                wellLocation=LiquidHandlingWellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=WellOffset(x=0, y=0, z=7.89),
                ),
            ),
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="dropTip",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
            ),
        ),
        protocol_schema_v7.Command(
            commandType="dropTip",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="dropTip",
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "wellName": "A1",
            },
        ),
        pe_commands.DropTipCreate(
            params=pe_commands.DropTipParams(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=DropTipWellLocation(
                    origin=DropTipWellOrigin.DEFAULT,
                    offset=WellOffset(x=0, y=0, z=0),
                ),
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="pickUpTip",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
            ),
        ),
        protocol_schema_v7.Command(
            commandType="pickUpTip",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="pickUpTip",
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "wellName": "A1",
            },
        ),
        pe_commands.PickUpTipCreate(
            params=pe_commands.PickUpTipParams(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=PickUpTipWellLocation(),
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="touchTip",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=-1.23),
                ),
            ),
        ),
        protocol_schema_v7.Command(
            commandType="touchTip",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=-1.23),
                ),
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="touchTip",
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "wellName": "A1",
                "wellLocation": {
                    "origin": "bottom",
                    "offset": {"x": 0, "y": 0, "z": -1.23},
                },
            },
        ),
        pe_commands.TouchTipCreate(
            params=pe_commands.TouchTipParams(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=WellOffset(x=0, y=0, z=-1.23),
                ),
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="loadPipette",
            params=protocol_schema_v6.Params(pipetteId="pipette-id-1", mount="left"),
        ),
        protocol_schema_v7.Command(
            commandType="loadPipette",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1", mount="left", pipetteName="p10_single"
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="loadPipette",
            params={
                "pipetteId": "pipette-id-1",
                "mount": "left",
                "pipetteName": "p10_single",
            },
        ),
        pe_commands.LoadPipetteCreate(
            params=pe_commands.LoadPipetteParams(
                pipetteId="pipette-id-1",
                pipetteName=PipetteNameType("p10_single"),
                mount=MountType("left"),
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="loadModule",
            params=protocol_schema_v6.Params(
                moduleId="module-id-1",
                location=Location(slotName="3"),
            ),
        ),
        protocol_schema_v7.Command(
            commandType="loadModule",
            params=protocol_schema_v7.Params(
                moduleId="module-id-1",
                model="magneticModuleV2",
                location=Location(slotName="3"),
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="loadModule",
            params={
                "moduleId": "module-id-1",
                "model": "magneticModuleV2",
                "location": {"slotName": "3"},
            },
        ),
        pe_commands.LoadModuleCreate(
            params=pe_commands.LoadModuleParams(
                model=ModuleModel("magneticModuleV2"),
                moduleId="module-id-1",
                location=DeckSlotLocation(slotName=(DeckSlotName("3"))),
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="loadLabware",
            params=protocol_schema_v6.Params(
                labwareId="labware-id-2",
                location=Location(moduleId="temperatureModuleId"),
            ),
        ),
        protocol_schema_v7.Command(
            commandType="loadLabware",
            params=protocol_schema_v7.Params(
                labwareId="labware-id-2",
                version=1,
                namespace="example",
                loadName="foo_8_plate_33ul",
                location=Location(moduleId="temperatureModuleId"),
                displayName="Trash",
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="loadLabware",
            params={
                "labwareId": "labware-id-2",
                "version": 1,
                "namespace": "example",
                "loadName": "foo_8_plate_33ul",
                "location": {"moduleId": "temperatureModuleId"},
                "displayName": "Trash",
            },
        ),
        pe_commands.LoadLabwareCreate(
            params=pe_commands.LoadLabwareParams(
                loadName="foo_8_plate_33ul",
                displayName="Trash",
                labwareId="labware-id-2",
                location=ModuleLocation(moduleId="temperatureModuleId"),
                version=1,
                namespace="example",
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="blowout",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
                flowRate=1.23,
            ),
        ),
        protocol_schema_v7.Command(
            commandType="blowout",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=SD_WellLocation(
                    origin="bottom",
                    offset=OffsetVector(x=0, y=0, z=7.89),
                ),
                flowRate=1.23,
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="blowout",
            params={
                "pipetteId": "pipette-id-1",
                "labwareId": "labware-id-2",
                "wellName": "A1",
                "wellLocation": {
                    "origin": "bottom",
                    "offset": {"x": 0, "y": 0, "z": 7.89},
                },
                "flowRate": 1.23,
            },
        ),
        pe_commands.BlowOutCreate(
            params=pe_commands.BlowOutParams(
                pipetteId="pipette-id-1",
                labwareId="labware-id-2",
                wellName="A1",
                wellLocation=WellLocation(
                    origin=WellOrigin.BOTTOM,
                    offset=WellOffset(x=0, y=0, z=7.89),
                ),
                flowRate=1.23,
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="delay",
            params=protocol_schema_v6.Params(waitForResume=True, message="hello world"),
        ),
        protocol_schema_v7.Command(
            commandType="delay",
            params=protocol_schema_v7.Params(waitForResume=True, message="hello world"),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="delay",
            params={"waitForResume": True, "message": "hello world"},
        ),
        pe_commands.WaitForResumeCreate(
            params=pe_commands.WaitForResumeParams(message="hello world")
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="delay",
            params=protocol_schema_v6.Params(seconds=12.34, message="hello world"),
        ),
        protocol_schema_v7.Command(
            commandType="delay",
            params=protocol_schema_v7.Params(seconds=12.34, message="hello world"),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="delay",
            params={"seconds": 12.34, "message": "hello world"},
        ),
        pe_commands.WaitForDurationCreate(
            params=pe_commands.WaitForDurationParams(
                seconds=12.34,
                message="hello world",
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="waitForResume",
            params=protocol_schema_v6.Params(message="hello world"),
        ),
        protocol_schema_v7.Command(
            commandType="waitForResume",
            params=protocol_schema_v7.Params(message="hello world"),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="waitForResume",
            params={"message": "hello world"},
        ),
        pe_commands.WaitForResumeCreate(
            params=pe_commands.WaitForResumeParams(message="hello world")
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="waitForDuration",
            params=protocol_schema_v6.Params(seconds=12.34, message="hello world"),
        ),
        protocol_schema_v7.Command(
            commandType="waitForDuration",
            params=protocol_schema_v7.Params(seconds=12.34, message="hello world"),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="waitForDuration",
            params={"seconds": 12.34, "message": "hello world"},
        ),
        pe_commands.WaitForDurationCreate(
            params=pe_commands.WaitForDurationParams(
                seconds=12.34,
                message="hello world",
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="moveToCoordinates",
            params=protocol_schema_v6.Params(
                pipetteId="pipette-id-1",
                coordinates=OffsetVector(x=1.1, y=2.2, z=3.3),
                minimumZHeight=123.4,
                forceDirect=True,
            ),
        ),
        protocol_schema_v7.Command(
            commandType="moveToCoordinates",
            params=protocol_schema_v7.Params(
                pipetteId="pipette-id-1",
                coordinates=OffsetVector(x=1.1, y=2.2, z=3.3),
                minimumZHeight=123.4,
                forceDirect=True,
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="moveToCoordinates",
            params={
                "pipetteId": "pipette-id-1",
                "coordinates": {"x": 1.1, "y": 2.2, "z": 3.3},
                "minimumZHeight": 123.4,
                "forceDirect": True,
            },
        ),
        pe_commands.MoveToCoordinatesCreate(
            params=pe_commands.MoveToCoordinatesParams(
                pipetteId="pipette-id-1",
                coordinates=DeckPoint(x=1.1, y=2.2, z=3.3),
                minimumZHeight=123.4,
                forceDirect=True,
            )
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="thermocycler/runProfile",
            params=protocol_schema_v6.Params(
                moduleId="module-id-2",
                blockMaxVolumeUl=1.11,
                profile=[
                    ProfileStep(
                        celsius=2.22,
                        holdSeconds=3.33,
                        rampRate=0.0,
                    ),
                    ProfileStep(
                        celsius=4.44,
                        holdSeconds=5.55,
                        rampRate=0.0,
                    ),
                ],
            ),
        ),
        protocol_schema_v7.Command(
            commandType="thermocycler/runProfile",
            params=protocol_schema_v7.Params(
                moduleId="module-id-2",
                blockMaxVolumeUl=1.11,
                profile=[
                    ProfileStep(
                        celsius=2.22,
                        holdSeconds=3.33,
                        rampRate=0.0,
                    ),
                    ProfileStep(
                        celsius=4.44,
                        holdSeconds=5.55,
                        rampRate=0.0,
                    ),
                ],
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="thermocycler/runProfile",
            params={
                "moduleId": "module-id-2",
                "blockMaxVolumeUl": 1.11,
                "profile": [
                    {
                        "celsius": 2.22,
                        "holdSeconds": 3.33,
                        "rampRate": 0.0,
                    },
                    {
                        "celsius": 4.44,
                        "holdSeconds": 5.55,
                        "rampRate": 0.0,
                    },
                ],
            },
        ),
        pe_commands.thermocycler.RunProfileCreate(
            params=pe_commands.thermocycler.RunProfileParams(
                moduleId="module-id-2",
                blockMaxVolumeUl=1.11,
                profile=[
                    pe_commands.thermocycler.RunProfileStepParams(
                        celsius=2.22, holdSeconds=3.33, rampRate=0.0
                    ),
                    pe_commands.thermocycler.RunProfileStepParams(
                        celsius=4.44, holdSeconds=5.55, rampRate=0.0
                    ),
                ],
            ),
        ),
    ),
    (
        protocol_schema_v6.Command(
            commandType="loadLiquid",
            key=None,
            params=protocol_schema_v6.Params(
                labwareId="labware-id-2",
                liquidId="liquid-id-555",
                volumeByWell={"A1": 32, "B2": 50},
            ),
        ),
        protocol_schema_v7.Command(
            commandType="loadLiquid",
            key=None,
            params=protocol_schema_v7.Params(
                labwareId="labware-id-2",
                liquidId="liquid-id-555",
                volumeByWell={"A1": 32, "B2": 50},
            ),
        ),
        protocol_schema_v8.Command.model_construct(
            commandType="loadLiquid",
            key=None,
            params={
                "labwareId": "labware-id-2",
                "liquidId": "liquid-id-555",
                "volumeByWell": {"A1": 32, "B2": 50},
            },
        ),
        pe_commands.LoadLiquidCreate(
            key=None,
            params=pe_commands.LoadLiquidParams(
                labwareId="labware-id-2",
                liquidId="liquid-id-555",
                volumeByWell={"A1": 32, "B2": 50},
            ),
        ),
    ),
]


@pytest.fixture
def subject() -> JsonTranslator:
    """Get a JsonTranslator test subject."""
    return JsonTranslator()


def _load_labware_definition_data() -> LabwareDefinition2:
    return LabwareDefinition2(
        version=1,
        namespace="example",
        schemaVersion=2,
        ordering=[["A1", "B1", "C1", "D1"], ["A2", "B2", "C2", "D2"]],
        groups=[Group(wells=["A1"], metadata=GroupMetadata())],
        wells={
            "A1": CircularWellDefinition2(
                depth=25,
                x=18.21,
                y=75.43,
                z=75,
                totalLiquidVolume=1100000,
                diameter=1,
                shape="circular",
            )
        },
        dimensions=Dimensions(yDimension=85.5, zDimension=100, xDimension=127.75),
        cornerOffsetFromSlot=SD_Labware_Vector(x=0, y=0, z=0),
        brand=BrandData(brand="foo"),
        metadata=Metadata(
            displayName="Foo 8 Well Plate 33uL",
            displayCategory=DisplayCategory("wellPlate"),
            displayVolumeUnits="µL",
        ),
        parameters=Parameters2(
            loadName="foo_8_plate_33ul",
            isTiprack=False,
            isMagneticModuleCompatible=False,
            format="irregular",
        ),
    )


def _make_v6_json_protocol(
    *,
    pipettes: Dict[str, Pipette] = {
        "pipette-id-1": Pipette(name="p10_single"),
    },
    labware_definitions: Dict[str, LabwareDefinition2] = {
        "example/plate/1": _load_labware_definition_data(),
        "example/trash/1": _load_labware_definition_data(),
    },
    labware: Dict[str, Labware] = {
        "labware-id-1": Labware(
            displayName="Source Plate", definitionId="example/plate/1"
        ),
        "labware-id-2": Labware(displayName="Trash", definitionId="example/trash/1"),
    },
    commands: List[protocol_schema_v6.Command] = [],
    modules: Dict[str, Module] = {
        "module-id-1": Module(model="magneticModuleV2"),
        "module-id-2": Module(model="thermocyclerModuleV2"),
    },
    liquids: Dict[str, Liquid] = {
        "liquid-id-555": Liquid(
            displayName="water", description="water description", displayColor="#F00"
        )
    },
) -> protocol_schema_v6.ProtocolSchemaV6:
    """Return a minimal JsonProtocol with the given elements, to use as test input."""
    return protocol_schema_v6.ProtocolSchemaV6(
        otSharedSchema="#/protocol/schemas/6",
        schemaVersion=6,
        metadata=SD_Metadata(),
        robot=Robot(model="OT-2 Standard", deckId="ot2_standard"),
        pipettes=pipettes,
        labwareDefinitions=labware_definitions,
        labware=labware,
        commands=commands,
        liquids=liquids,
        modules=modules,
    )


def _make_v7_json_protocol(
    *,
    labware_definitions: Dict[str, LabwareDefinition2] = {
        "example/plate/1": _load_labware_definition_data(),
        "example/trash/1": _load_labware_definition_data(),
    },
    commands: List[protocol_schema_v7.Command] = [],
    liquids: Dict[str, Liquid] = {
        "liquid-id-555": Liquid(
            displayName="water", description="water description", displayColor="#F00"
        )
    },
) -> protocol_schema_v7.ProtocolSchemaV7:
    """Return a minimal JsonProtocol with the given elements, to use as test input."""
    return protocol_schema_v7.ProtocolSchemaV7(
        otSharedSchema="#/protocol/schemas/7",
        schemaVersion=7,
        metadata=SD_Metadata(),
        robot=Robot(model="OT-2 Standard", deckId="ot2_standard"),
        labwareDefinitions=labware_definitions,
        commands=commands,
        liquids=liquids,
    )


def _make_v8_json_protocol(
    *,
    labware_definitions: Dict[str, LabwareDefinition] = {
        "example/plate/1": _load_labware_definition_data(),
        "example/trash/1": _load_labware_definition_data(),
    },
    commands: List[protocol_schema_v8.Command] = [],
    liquids: Dict[str, Liquid] = {
        "liquid-id-555": Liquid(
            displayName="water", description="water description", displayColor="#F00"
        )
    },
) -> protocol_schema_v8.ProtocolSchemaV8:
    """Return a minimal json protocol with the given elements."""
    return protocol_schema_v8.ProtocolSchemaV8(
        otSharedSchema="#/protocol/schemas/8",
        schemaVersion=8,
        metadata=SD_Metadata(),
        robot=Robot(
            model="OT-2 Standard",
            deckId="ot2_standard",
        ),
        labwareDefinitions=labware_definitions,
        labwareDefinitionSchemaId="opentronsLabwareSchemaV2",
        commandSchemaId=protocol_schema_v8.CommandSchemaId("opentronsCommandSchemaV8"),
        commands=commands,
        liquidSchemaId="opentronsLiquidSchemaV1",
        liquids=liquids,
        commandAnnotationSchemaId="opentronsCommandAnnotationSchemaV1",
        commandAnnotations=[],
    )


@pytest.mark.parametrize(
    "test_v6_input,test_v7_input,test_v8_input,expected_output", VALID_TEST_PARAMS
)
def test_load_command(
    subject: JsonTranslator,
    test_v6_input: protocol_schema_v6.Command,
    test_v7_input: protocol_schema_v7.Command,
    test_v8_input: protocol_schema_v8.Command,
    expected_output: pe_commands.CommandCreate,
) -> None:
    """Test translating v6 commands to protocol engine commands."""
    v6_output = subject.translate_commands(
        _make_v6_json_protocol(commands=[test_v6_input])
    )
    v7_output = subject.translate_commands(
        _make_v7_json_protocol(commands=[test_v7_input])
    )
    v8_output = subject.translate_commands(
        _make_v8_json_protocol(commands=[test_v8_input])
    )
    assert v6_output == [expected_output]
    assert v7_output == [expected_output]
    assert v8_output == [expected_output]


def test_load_liquid(
    subject: JsonTranslator,
) -> None:
    """Test translating v6 commands to protocol engine commands."""
    protocol = _make_v6_json_protocol()
    result = subject.translate_liquids(protocol)

    assert result == [
        PE_Liquid(
            id="liquid-id-555",
            displayName="water",
            description="water description",
            displayColor=HexColor("#F00"),
        )
    ]



from typing import AsyncGenerator, Type, Union

import mock
import pytest
from _pytest.fixtures import SubRequest
from mock import AsyncMock, call

from opentrons.drivers.asyncio.communication.async_serial import AsyncSerial
from opentrons.drivers.asyncio.communication.errors import (
    AlarmResponse,
    BaseErrorCode,
    DefaultErrorCodes,
    ErrorResponse,
    NoResponse,
    UnhandledGcode,
)
from opentrons.drivers.asyncio.communication.serial_connection import (
    AsyncResponseSerialConnection,
    SerialConnection,
)


@pytest.fixture
def mock_serial_port() -> AsyncMock:
    return AsyncMock(spec=AsyncSerial)


@pytest.fixture
def ack() -> str:
    return "ack"


SerialKind = Union[AsyncResponseSerialConnection, SerialConnection]


# Async because SerialConnection.__init__() needs an event loop,
# so this fixture needs to run in an event loop.
@pytest.fixture(
    params=[AsyncResponseSerialConnection, SerialConnection],
)
async def subject(
    request: SubRequest, mock_serial_port: AsyncMock, ack: str
) -> SerialKind:
    """Create the test subject."""
    serial_class = request.param
    serial_class.RETRY_WAIT_TIME = 0
    if serial_class == AsyncResponseSerialConnection:
        return serial_class(  # type: ignore[no-any-return]
            serial=mock_serial_port,
            ack=ack,
            name="name",
            port="port",
            retry_wait_time_seconds=0,
            error_keyword="err",
            alarm_keyword="alarm",
            async_error_ack="async",
            error_codes=DefaultErrorCodes,
        )
    elif serial_class == SerialConnection:
        return serial_class(  # type: ignore[no-any-return]
            serial=mock_serial_port,
            ack=ack,
            name="name",
            port="port",
            retry_wait_time_seconds=0,
            error_keyword="error",
            alarm_keyword="alarm",
        )
    else:
        raise AssertionError(f"Unexpected serial_class: {serial_class}")


@pytest.fixture
async def async_subject(
    mock_serial_port: AsyncMock, ack: str
) -> AsyncResponseSerialConnection:
    """Create the test async subject."""
    AsyncResponseSerialConnection.RETRY_WAIT_TIME = 0  # type: ignore[attr-defined]
    return AsyncResponseSerialConnection(
        serial=mock_serial_port,
        ack=ack,
        name="name",
        port="port",
        retry_wait_time_seconds=0,
        error_keyword="err",
        alarm_keyword="alarm",
        async_error_ack="async",
    )


@pytest.fixture
async def subject_raise_on_error_patched(
    async_subject: AsyncResponseSerialConnection,
) -> AsyncGenerator[AsyncResponseSerialConnection, None]:
    raise_on_error_mock = mock.MagicMock()
    with mock.patch.object(async_subject, "raise_on_error", raise_on_error_mock):
        yield async_subject


async def test_send_command(
    mock_serial_port: AsyncMock, subject: SerialKind, ack: str
) -> None:
    """It should send a command."""
    serial_response = "response data " + ack
    mock_serial_port.read_until.return_value = serial_response.encode()

    await subject.send_data(data="send data")

    mock_serial_port.timeout_override.assert_called_once_with("timeout", None)
    mock_serial_port.write.assert_called_once_with(data=b"send data")
    mock_serial_port.read_until.assert_called_once_with(match=ack.encode())


async def test_send_command_with_retry(
    mock_serial_port: AsyncMock, subject: SerialKind, ack: str
) -> None:
    """It should retry sending after a read failure."""
    serial_response = "response data " + ack
    mock_serial_port.read_until.side_effect = (b"", serial_response.encode())

    await subject.send_data(data="send data", retries=1)

    mock_serial_port.timeout_override.assert_called_once_with("timeout", None)
    mock_serial_port.write.assert_has_calls(
        calls=[call(data=b"send data"), call(data=b"send data")]
    )
    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
            call(match=ack.encode()),
        ]
    )


async def test_send_command_with_zero_retries(
    mock_serial_port: AsyncMock, async_subject: AsyncResponseSerialConnection, ack: str
) -> None:
    """It should a command once"""
    mock_serial_port.read_until.side_effect = (b"", b"")

    # Set the default number of retries to 1, we want to overide this with
    # the retries from the subject.send_data(data, retries=0) method call.
    async_subject._number_of_retries = 1

    with pytest.raises(NoResponse):
        # We want this to overwrite the internal `_number_of_retries`
        await async_subject.send_data(data="send data", retries=0)

    mock_serial_port.timeout_override.assert_called_once_with("timeout", None)
    mock_serial_port.write.assert_called_once_with(data=b"send data")
    mock_serial_port.read_until.assert_called_once_with(match=ack.encode())
    mock_serial_port.close.assert_called_once()
    mock_serial_port.open.assert_called_once()


async def test_send_command_with_retry_exhausted(
    mock_serial_port: AsyncMock, subject: SerialKind
) -> None:
    """It should raise after retries exhausted."""
    mock_serial_port.read_until.side_effect = (b"", b"", b"")

    with pytest.raises(NoResponse):
        await subject.send_data(data="send data", retries=2)


async def test_send_command_response(
    mock_serial_port: AsyncMock, subject: SerialKind, ack: str
) -> None:
    """It should return response without the ack and stripped."""
    response_data = "response data"
    serial_response = f" {response_data}  {ack}"
    mock_serial_port.read_until.return_value = serial_response.encode()

    response = await subject.send_data(data="send data")

    assert response == response_data


@pytest.mark.parametrize(
    argnames=["response", "exception_type", "async_only"],
    argvalues=[
        ["error", ErrorResponse, False],
        ["Error", ErrorResponse, False],
        ["Error: was found.", ErrorResponse, False],
        ["alarm", AlarmResponse, False],
        ["ALARM", AlarmResponse, False],
        ["This is an Alarm", AlarmResponse, False],
        ["error:Alarm lock", AlarmResponse, False],
        ["alarm:error", AlarmResponse, False],
        ["ALARM: Hard limit -X", AlarmResponse, False],
        ["ERR003:unhandled gcode OK ", UnhandledGcode, True],
    ],
)
def test_raise_on_error(
    subject: SerialKind,
    response: str,
    exception_type: Type[Exception],
    async_only: bool,
) -> None:
    """It should raise an exception on error/alarm responses."""
    if isinstance(subject, SerialConnection) and async_only:
        pytest.skip()
    with pytest.raises(expected_exception=exception_type, match=response):
        subject.raise_on_error(response, "fake request")


def test_raise_on_error_no_raise_on_keyword_in_body(
    subject: SerialKind,
) -> None:
    """It should not raise when there is a keyword in the response body."""
    request = "M226 Z"
    # This response contains `eRR` which tricks the system into thinking there is an
    # error, we fixed this by making sure the request and response gcodes match.
    response = "M226 Z I:12 D:gW2ACQuAAAAAAAAAAAAAAAAAAAABAQAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAACPbxeRRikcFhINCQYFBAICAQEBAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    subject.raise_on_error(response, request)

    # This should still raise
    with pytest.raises(expected_exception=ErrorResponse, match="error"):
        subject.raise_on_error("error", request)


def test_get_error_codes_lowercase(
    subject: SerialKind,
) -> None:
    """It should return an error code dictionary keyed by lowercase value."""
    lowercase_result = subject._error_codes.get_error_codes()
    assert lowercase_result == {"err003": DefaultErrorCodes.UNHANDLED_GCODE}


async def test_on_retry(mock_serial_port: AsyncMock, subject: SerialKind) -> None:
    """It should try to re-open connection."""
    await subject.on_retry()

    mock_serial_port.close.assert_called_once()
    mock_serial_port.open.assert_called_once()


async def test_send_data_with_async_error_before(
    mock_serial_port: AsyncMock,
    subject_raise_on_error_patched: AsyncResponseSerialConnection,
    ack: str,
) -> None:
    """It should return response without the ack and stripped. It should also handle the async error."""
    error_response = "async ERR106:main motor:speedsensor failed"
    serial_error_response = f" {error_response}  {ack}"
    encoded_error_response = serial_error_response.encode()
    successful_response = "G28"
    data = "G28"
    serial_successful_response = f" {successful_response}  {ack}"
    encoded_successful_response = serial_successful_response.encode()
    mock_serial_port.read_until.side_effect = [
        encoded_error_response,
        encoded_successful_response,
    ]

    response = await subject_raise_on_error_patched._send_data(data=data, retries=0)

    assert response == successful_response
    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
            call(match=ack.encode()),
        ]
    )
    subject_raise_on_error_patched.raise_on_error.assert_has_calls(  # type: ignore[attr-defined]
        calls=[
            call(response=error_response, request=data),
            call(response=successful_response, request=data),
        ]
    )


async def test_send_data_with_async_error_after(
    mock_serial_port: AsyncMock,
    subject_raise_on_error_patched: AsyncResponseSerialConnection,
    ack: str,
) -> None:
    """It should return response without the ack and stripped. It should not handle the async error."""
    error_response = "async ERR106:main motor:speedsensor failed"
    serial_error_response = f" {error_response}  {ack}"
    encoded_error_response = serial_error_response.encode()
    successful_response = "G28"
    data = "G28"
    serial_successful_response = f" {successful_response}  {ack}"
    encoded_successful_response = serial_successful_response.encode()
    mock_serial_port.read_until.side_effect = [
        encoded_successful_response,
        encoded_error_response,
    ]

    response = await subject_raise_on_error_patched._send_data(data=data, retries=0)

    assert response == successful_response
    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
        ]
    )
    subject_raise_on_error_patched.raise_on_error.assert_has_calls(  # type: ignore[attr-defined]
        calls=[
            call(response=successful_response, request=data),
        ]
    )


async def test_send_data_multiple_ack_ok(
    mock_serial_port: AsyncMock,
    async_subject: AsyncResponseSerialConnection,
    ack: str,
) -> None:
    """It should return all acks."""
    successful_response = "M411"
    data = "M411"
    serial_successful_response = f" {successful_response}  {ack}"
    encoded_successful_response = serial_successful_response.encode()
    mock_serial_port.read_until.side_effect = [
        encoded_successful_response,
        encoded_successful_response,
        encoded_successful_response,
    ]

    responses = await async_subject._send_data_multiack(data=data, retries=0, acks=3)

    assert responses == [successful_response] * 3
    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
            call(match=ack.encode()),
            call(match=ack.encode()),
        ]
    )


async def test_send_data_multiple_ack_some_errors(
    mock_serial_port: AsyncMock,
    async_subject: AsyncResponseSerialConnection,
    ack: str,
) -> None:
    """It should return all acks."""
    successful_response = "M411"
    data = "M411"
    error_response = "ERR007:test"
    serial_successful_response = f" {successful_response}  {ack}"
    encoded_successful_response = serial_successful_response.encode()
    serial_error_response = f" {error_response}  {ack}"
    encoded_error_response = serial_error_response.encode()
    mock_serial_port.read_until.side_effect = [
        encoded_successful_response,
        encoded_error_response,
        encoded_successful_response,
    ]

    with pytest.raises(ErrorResponse, match=error_response):
        await async_subject._send_data_multiack(data=data, retries=0, acks=3)

    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
            call(match=ack.encode()),
            call(match=ack.encode()),
        ]
    )


async def test_send_data_multiple_ack_ok_with_async_error(
    mock_serial_port: AsyncMock,
    async_subject: AsyncResponseSerialConnection,
    ack: str,
) -> None:
    """It should return all acks."""
    successful_response = "M411"
    data = "M411"
    serial_successful_response = f" {successful_response}  {ack}"
    encoded_successful_response = serial_successful_response.encode()
    error_response = "async ERR106:main motor:speedsensor failed"
    serial_error_response = f" {error_response}  {ack}"
    encoded_error_response = serial_error_response.encode()
    mock_serial_port.read_until.side_effect = [
        encoded_error_response,
        encoded_successful_response,
        encoded_successful_response,
        encoded_successful_response,
    ]

    with pytest.raises(ErrorResponse, match=error_response):
        await async_subject._send_data_multiack(data=data, retries=0, acks=3)

    mock_serial_port.read_until.assert_has_calls(
        calls=[
            call(match=ack.encode()),
            call(match=ack.encode()),
            call(match=ack.encode()),
            call(match=ack.encode()),
        ]
    )


def test_default_error_code_raise_exception() -> None:
    """Test that error codes can raise appropriate exceptions."""
    with pytest.raises(UnhandledGcode) as error:
        DefaultErrorCodes.UNHANDLED_GCODE.raise_exception(
            port="test_port", response="ERR003:test", command="G28"
        )

    assert error.value.response == "ERR003:test"
    assert error.value.port == "test_port"
    assert error.value.command == "G28"


def test_custom_error_code_raise_custom_exception() -> None:
    """Test that custom error codes can raise appropriate exceptions."""

    class CustomErrorResponse(ErrorResponse):
        pass

    class CustomDefaultErrorCodes(BaseErrorCode):
        CUSTOM_ERROR = ("ERR999", CustomErrorResponse)

    # Test that a regular ErrorResponse works correctly
    with pytest.raises(CustomErrorResponse) as error:
        CustomDefaultErrorCodes.CUSTOM_ERROR.raise_exception(
            port="test_port", response="ERR999:test", command="G28"
        )

    assert error.value.command == "G28"
    assert error.value.response == "ERR999:test"
    assert error.value.port == "test_port"


async def test_send_data_multiple_raises_unhandled(
    mock_serial_port: AsyncMock, async_subject: AsyncResponseSerialConnection, ack: str
) -> None:
    """It shouldn't wait for both acks before raising an unhandled gcode"""
    mock_serial_port.read_until.side_effect = [
        f"ERR003:unhandled gcode {ack}".encode(),
    ]
    with pytest.raises(UnhandledGcode):
        await async_subject._send_data_multiack(data="M411", retries=0, acks=3)
    mock_serial_port.read_until.assert_called_once()



"""Test magnetic module core."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import MagDeck
from opentrons.hardware_control.modules.types import MagneticStatus, ModuleType
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine.exceptions import InvalidMagnetEngageHeightError
from opentrons.protocol_api.core.engine.module_core import MagneticModuleCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors.exceptions import (
    LabwareNotLoadedOnModuleError,
    NoMagnetEngageHeightError,
)
from opentrons.protocol_engine.types import ModuleModel

MagDeckHardware = SynchronousAdapter[MagDeck]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> MagDeckHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="MagDeckHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: MagDeckHardware,
    mock_protocol_core: ProtocolCore,
) -> MagneticModuleCore:
    """Get a mock of MagneticModuleCore."""
    return MagneticModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: MagDeckHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create a magnetic module core."""
    result = MagneticModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.MAGNETIC


def test_engage_from_base(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should verify a call to sync client engage method."""
    decoy.when(
        mock_engine_client.state.modules.get_connected_model(module_id="1234")
    ).then_return(ModuleModel.MAGNETIC_MODULE_V1)

    decoy.when(
        mock_engine_client.state.modules.calculate_magnet_height(
            module_model=ModuleModel.MAGNETIC_MODULE_V1, height_from_base=7.0
        )
    ).then_return(9.0)
    subject.engage(height_from_base=7.0)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.magnetic_module.EngageParams(moduleId="1234", height=7.0)
        ),
        times=1,
    )


def test_engage_to_labware(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should verify a call to sync client engage method."""
    decoy.when(
        mock_engine_client.state.modules.get_connected_model(module_id="1234")
    ).then_return(ModuleModel.MAGNETIC_MODULE_V1)

    decoy.when(
        mock_engine_client.state.labware.get_default_magnet_height(
            module_id="1234", offset=2.0
        )
    ).then_return(6.0)

    subject.engage_to_labware(offset=2.0)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.magnetic_module.EngageParams(moduleId="1234", height=6.0)
        ),
        times=1,
    )


def test_engage_to_labware_raises_no_labware_error(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should raise an error that the labware was not found."""
    decoy.when(
        mock_engine_client.state.labware.get_default_magnet_height(
            module_id="1234", offset=0
        )
    ).then_raise(
        LabwareNotLoadedOnModuleError  # type: ignore[arg-type]
    )

    with pytest.raises(InvalidMagnetEngageHeightError, match="no labware loaded"):
        subject.engage_to_labware(offset=0)


def test_engage_to_labware_raises_no_height_error(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should raise an error that the a magnetic default height was not found."""
    decoy.when(
        mock_engine_client.state.labware.get_default_magnet_height(
            module_id="1234", offset=0
        )
    ).then_raise(
        NoMagnetEngageHeightError  # type: ignore[arg-type]
    )

    with pytest.raises(
        InvalidMagnetEngageHeightError, match="does not have a default engage height"
    ):
        subject.engage_to_labware(offset=0)


def test_disengage(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should verify a call to sync client disengage method."""
    subject.disengage()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.magnetic_module.DisengageParams(moduleId="1234")
        ),
        times=1,
    )


def test_get_status(
    decoy: Decoy,
    subject: MagneticModuleCore,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: MagDeckHardware,
) -> None:
    """Should get the magnetic module status."""
    decoy.when(mock_sync_module_hardware.status).then_return(MagneticStatus.ENGAGED)

    assert subject.get_status() == MagneticStatus.ENGAGED


def test_get_serial_number(
    decoy: Decoy, subject: MagneticModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return a serial number."""
    decoy.when(mock_engine_client.state.modules.get_serial_number("1234")).then_return(
        "abc"
    )

    assert subject.get_serial_number() == "abc"



"""Tests for the Protocol Context's synchronous engine adapter.

Since Python protocol execution happens off the main thread, these tests call
the subject's methods in a synchronous context in a child thread to ensure:

- In the Protocol execution thread, calls are synchronous and block until
    command execution is complete.
- In the main thread, the Protocol Engine does its work in the main event
    loop, without blocking.
"""

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2
from opentrons_shared_data.labware.types import LabwareUri

from opentrons.protocol_engine import commands
from opentrons.protocol_engine.clients import ChildThreadTransport, SyncClient
from opentrons.protocol_engine.types import DeckPoint, DeckSlotLocation, Liquid
from opentrons.types import DeckSlotName


@pytest.fixture
def transport(decoy: Decoy) -> ChildThreadTransport:
    """Get a stubbed out ChildThreadTransport."""
    return decoy.mock(cls=ChildThreadTransport)


@pytest.fixture
def subject(transport: ChildThreadTransport) -> SyncClient:
    """Get a SyncProtocolEngine test subject."""
    return SyncClient(transport=transport)


def test_execute_command(
    decoy: Decoy, transport: ChildThreadTransport, subject: SyncClient
) -> None:
    """It should map the command params and execute it."""
    params = commands.CommentParams(message="hewwo")
    expected_request = commands.CommentCreate(params=params)
    subject.execute_command(params)
    decoy.verify(transport.execute_command_wait_for_recovery(request=expected_request))


def test_execute_command_without_recovery(
    decoy: Decoy, transport: ChildThreadTransport, subject: SyncClient
) -> None:
    """It should map the command params and execute it."""
    params = commands.LoadLabwareParams(
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A1),
        loadName="loadName",
        namespace="namespace",
        version=0,
    )
    expected_request = commands.LoadLabwareCreate(params=params)
    result_from_transport = commands.AspirateResult(
        position=DeckPoint(x=1, y=2, z=3), volume=4
    )
    decoy.when(transport.execute_command(expected_request)).then_return(
        result_from_transport
    )
    result_from_subject = subject.execute_command_without_recovery(params)
    assert result_from_subject == result_from_transport  # type: ignore[comparison-overlap]


def test_add_labware_definition(
    decoy: Decoy,
    transport: ChildThreadTransport,
    subject: SyncClient,
) -> None:
    """It should add a labware definition."""
    labware_definition = LabwareDefinition2.model_construct(namespace="hello")  # type: ignore[call-arg]
    expected_labware_uri = LabwareUri("hello/world/123")

    decoy.when(
        transport.call_method(
            "add_labware_definition",
            definition=labware_definition,
        )
    ).then_return(expected_labware_uri)

    result = subject.add_labware_definition(labware_definition)

    assert result == expected_labware_uri


def test_add_addressable_area(
    decoy: Decoy,
    transport: ChildThreadTransport,
    subject: SyncClient,
) -> None:
    """It should add an addressable area."""
    subject.add_addressable_area(addressable_area_name="cool-area")

    decoy.verify(
        transport.call_method(
            "add_addressable_area",
            addressable_area_name="cool-area",
        ),
        times=1,
    )


def test_add_liquid(
    decoy: Decoy,
    transport: ChildThreadTransport,
    subject: SyncClient,
) -> None:
    """It should add a liquid to engine state."""
    liquid = Liquid.model_construct(displayName="water")  # type: ignore[call-arg]

    decoy.when(
        transport.call_method(
            "add_liquid",
            name="water",
            description="water desc",
            color="#fff",
        )
    ).then_return(liquid)

    result = subject.add_liquid(name="water", description="water desc", color="#fff")

    assert result == liquid



"""Command models for Flex Stacker commands."""

from .empty import Empty, EmptyCommandType, EmptyCreate, EmptyParams, EmptyResult
from .fill import Fill, FillCommandType, FillCreate, FillParams, FillResult
from .retrieve import (
    Retrieve,
    RetrieveCommandType,
    RetrieveCreate,
    RetrieveParams,
    RetrieveResult,
)
from .set_stored_labware import (
    SetStoredLabware,
    SetStoredLabwareCommandType,
    SetStoredLabwareCreate,
    SetStoredLabwareParams,
    SetStoredLabwareResult,
    StackerStoredLabwareDetails,
)
from .store import (
    Store,
    StoreCommandType,
    StoreCreate,
    StoreParams,
    StoreResult,
)

__all__ = [
    # flexStacker/store
    "StoreCommandType",
    "StoreParams",
    "StoreResult",
    "Store",
    "StoreCreate",
    # flexStacker/retrieve
    "RetrieveCommandType",
    "RetrieveParams",
    "RetrieveResult",
    "Retrieve",
    "RetrieveCreate",
    # flexStacker/setStoredLabware
    "SetStoredLabwareCommandType",
    "SetStoredLabwareParams",
    "SetStoredLabwareResult",
    "SetStoredLabware",
    "SetStoredLabwareCreate",
    "StackerStoredLabwareDetails",
    # flexStacker/fill
    "FillCommandType",
    "FillParams",
    "FillResult",
    "Fill",
    "FillCreate",
    # flexStacker/empty
    "EmptyCommandType",
    "EmptyParams",
    "EmptyResult",
    "Empty",
    "EmptyCreate",
]



from __future__ import annotations

import asyncio
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Mapping,
    Optional,
)

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.vacuum_module.abstract import AbstractVacuumModuleDriver
from opentrons.drivers.vacuum_module.driver import (
    VacuumModuleDriver,
)
from opentrons.drivers.vacuum_module.simulator import SimulatingDriver
from opentrons.drivers.vacuum_module.types import (
    LEDColor,
    LEDPattern,
)
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.modules import mod_abc, update
from opentrons.hardware_control.modules.types import (
    LiveData,
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
    ModuleType,
    UploadFunction,
    VacuumModuleData,
    VacuumModuleStatus,
)
from opentrons.hardware_control.poller import Poller, Reader
from opentrons.hardware_control.types import StatusBarState, StatusBarUpdateEvent

log = logging.getLogger(__name__)

POLL_PERIOD = 2.0
SIMULATING_POLL_PERIOD = POLL_PERIOD / 20.0

DFU_PID = "df11"


class VacuumModule(mod_abc.AbstractModule):
    """Hardware control interface for an attached Flex-Stacker module."""

    MODULE_TYPE = ModuleType.VACUUM_MODULE

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        hw_control_loop: asyncio.AbstractEventLoop,
        execution_manager: ExecutionManager,
        disconnected_callback: ModuleDisconnectedCallback,
        error_callback: ModuleErrorCallback,
        poll_interval_seconds: float | None = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
    ) -> "VacuumModule":
        """
        Build a VacuumModule

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            VacuumModule instance
        """
        driver: AbstractVacuumModuleDriver
        if not simulating:
            driver = await VacuumModuleDriver.create(port=port, loop=hw_control_loop)
            poll_interval_seconds = poll_interval_seconds or POLL_PERIOD
        else:
            driver = SimulatingDriver(serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIMULATING_POLL_PERIOD

        reader = VacuumModuleReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=(await driver.get_device_info()).to_dict(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
            error_callback=error_callback,
        )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Flex-Stacker on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        driver: AbstractVacuumModuleDriver,
        reader: VacuumModuleReader,
        poller: Poller,
        device_info: Mapping[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        execution_manager: ExecutionManager,
        disconnected_callback: ModuleDisconnectedCallback,
        error_callback: ModuleErrorCallback,
    ):
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
            error_callback=error_callback,
        )
        self._device_info = device_info
        self._driver = driver
        self._reader = reader
        self._poller = poller
        self._last_status_bar_event: Optional[StatusBarUpdateEvent] = None
        self._should_identify = False
        self._device_status = VacuumModuleStatus.IDLE
        # Set initialized callback
        self._unsubscribe_init = reader.set_initialized_callback(
            self._initialized_callback
        )
        self._unsubscribe_error = reader.set_error_callback(self._async_error_callback)

    async def _initialized_callback(self) -> None:
        """Called by the reader once the module is initialized."""
        if self._last_status_bar_event:
            await self._handle_status_bar_event(self._last_status_bar_event)

    def _async_error_callback(self, exception: Exception) -> None:
        self.error_callback(exception)

    async def cleanup(self) -> None:
        """Stop the poller task"""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        """Used for picking up serial port symlinks"""
        return "vacuummodule"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "vacuum-module"

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        return "vacuumModuleV1"

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    @property
    def initialized(self) -> bool:
        """The stacker is ready..."""
        return self._reader.initialized

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def status(self) -> VacuumModuleStatus:
        """Module status or error state details."""
        return self._device_status

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    @property
    def live_data(self) -> LiveData:
        data: VacuumModuleData = {
            "errorDetails": self._reader.error,
        }
        return {"status": self.status.value, "data": data}

    @property
    def should_identify(self) -> bool:
        return self._should_identify

    async def prep_for_update(self) -> str:
        await self._poller.stop()
        await self._driver.set_vacuum_state(False)
        await self._driver.enter_programming_mode()
        # This device has three unique "devices" over DFU
        dfu_info = await update.find_dfu_device(pid=DFU_PID, expected_device_count=3)
        return dfu_info

    def bootloader(self) -> UploadFunction:
        return update.upload_via_dfu

    async def deactivate(self, must_be_running: bool = True) -> None:
        pass

    async def set_led_state(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Sets the statusbar state."""
        return await self._driver.set_led(
            power, color=color, pattern=pattern, duration=duration, reps=reps
        )

    def event_listener(self, event: Any) -> None:
        if isinstance(event, StatusBarUpdateEvent):
            self._last_status_bar_event = event
            asyncio.run_coroutine_threadsafe(
                self._handle_status_bar_event(event), self._loop
            )

    async def _handle_status_bar_event(self, event: StatusBarUpdateEvent) -> None:  # noqa: C901
        if event.enabled and self.initialized:
            match event.state:
                case StatusBarState.RUNNING:
                    await self.set_led_state(0.5, LEDColor.GREEN, LEDPattern.STATIC)
                case StatusBarState.PAUSED:
                    if self.should_identify:
                        await self._statusbar_pause()
                    else:
                        await self._statusbar_idle()
                case StatusBarState.HARDWARE_ERROR:
                    if self.should_identify:
                        await self.set_led_state(
                            0.5, LEDColor.RED, LEDPattern.FLASH, duration=300
                        )
                    else:
                        await self._statusbar_idle()
                case StatusBarState.SOFTWARE_ERROR:
                    await self.set_led_state(0.5, LEDColor.YELLOW, LEDPattern.STATIC)
                case StatusBarState.ERROR_RECOVERY:
                    if self.should_identify:
                        await self.set_led_state(
                            0.5, LEDColor.YELLOW, LEDPattern.PULSE, duration=2000
                        )
                    else:
                        await self._statusbar_idle()
                case StatusBarState.RUN_COMPLETED:
                    await self.set_led_state(0.5, LEDColor.GREEN, LEDPattern.PULSE)
                case StatusBarState.UPDATING:
                    await self.set_led_state(0.5, LEDColor.WHITE, LEDPattern.PULSE)
                case StatusBarState.IDLE | _:
                    await self._statusbar_idle()

    async def _statusbar_pause(self) -> None:
        await self.set_led_state(0.5, LEDColor.BLUE, LEDPattern.PULSE, duration=2000)

    async def _statusbar_idle(self) -> None:
        await self.set_led_state(0.5, LEDColor.WHITE, LEDPattern.STATIC)

    async def identify(self, start: bool, color_name: Optional[str] = None) -> None:
        """Identify the module."""
        reps = -1 if start else 0
        color = LEDColor.from_name(color_name or LEDColor.BLUE.name)
        await self.set_led_state(0.5, color, LEDPattern.PULSE, reps=reps)
        if not start and self._last_status_bar_event:
            await self._handle_status_bar_event(self._last_status_bar_event)

    def set_statusbar_identify(self, state: bool) -> None:
        self._should_identify = state

    def cleanup_persistent(self) -> None:
        """Reset persistent data on the module that should not exist outside of a run."""
        self.set_statusbar_identify(False)


class VacuumModuleReader(Reader):
    error: Optional[str]

    def __init__(self, driver: AbstractVacuumModuleDriver) -> None:
        self.error: Optional[str] = None
        self._driver = driver
        self.initialized = False
        self._refresh_state = False
        self._initialized_callback: Optional[Callable[[], Awaitable[None]]] = None
        self._error_callback: Optional[Callable[[Exception], None]] = None

    def set_initialized_callback(
        self, callback: Callable[[], Awaitable[None]]
    ) -> Callable[[], None]:
        """Sets the callback used when done initializing the module."""
        self._initialized_callback = callback
        return self._remove_init_callback

    def _remove_init_callback(self) -> None:
        self._initialized_callback = None

    def set_error_callback(
        self, error_callback: Callable[[Exception], None]
    ) -> Callable[[], None]:
        """Register a handler for asynchronous hardware errors."""
        self._error_callback = error_callback
        return self._remove_error_callback

    def _remove_error_callback(self) -> None:
        self._error_callback = None

    async def read(self) -> None:
        if not self.initialized or self._refresh_state:
            initialized = True
            self._refresh_state = False
            # We are done initializing, sync the led state
            if not self.initialized and initialized:
                self.initialized = True
                if self._initialized_callback:
                    await self._initialized_callback()

        self._set_error(None)

    def set_refresh_state(self) -> None:
        """Tell the reader to refresh all states, even ones that arent polled."""
        self._refresh_state = True

    def on_error(self, exception: Exception) -> None:
        self._driver.reset_serial_buffers()
        self._set_error(exception)

    def _set_error(self, exception: Optional[Exception]) -> None:
        if exception is None:
            self.error = None
        else:
            if self._error_callback:
                self._error_callback(exception)
            try:
                self.error = str(exception.args[0])
            except Exception:
                self.error = repr(exception)



import importlib
import json
from types import ModuleType
from typing import Generator, no_type_check

import pytest
from pydantic import ValidationError

from opentrons.calibration_storage import encoder_decoder as ed
from opentrons.calibration_storage import types as cal_types
from opentrons.util.helpers import utc_now


@no_type_check
@pytest.fixture
def model(request: pytest.FixtureRequest) -> Generator[ModuleType, None, None]:
    robot_type = request.param
    if robot_type == "ot3":
        yield importlib.import_module("opentrons.calibration_storage.ot3.models")
    else:
        yield importlib.import_module("opentrons.calibration_storage.ot2.models")


def test_json_datetime_encoder() -> None:
    fake_time = utc_now()
    original = {"mock_hash": {"tipLength": 25.0, "lastModified": fake_time}}

    encoded = json.dumps(original, cls=ed.DateTimeEncoder)
    decoded = json.loads(encoded, cls=ed.DateTimeDecoder)
    assert decoded == original


@no_type_check
@pytest.mark.parametrize(
    argnames=["model"],
    argvalues=[["ot2"], ["ot3"]],
    indirect=True,
)
def test_tip_length_positive(model: ModuleType) -> None:
    with pytest.raises(ValidationError):
        model.v1.TipLengthModel(
            tipLength=-10,
            lastModified=utc_now(),
            source=cal_types.SourceType.user,
            status=model.v1.CalibrationStatus(),
            uri="opentrons/tiprack/1",
        )



"""Test absorbance reader initilize command."""

from typing import List

import pytest
from decoy import Decoy

from opentrons.drivers.types import ABSMeasurementMode
from opentrons.hardware_control.modules import AbsorbanceReader
from opentrons.protocol_engine.commands.absorbance_reader import (
    InitializeParams,
    InitializeResult,
)
from opentrons.protocol_engine.commands.absorbance_reader.initialize import (
    InitializeImpl,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors import InvalidWavelengthError
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates import (
    AbsorbanceReaderId,
    AbsorbanceReaderSubState,
)
from opentrons.protocol_engine.state.state import StateView


@pytest.fixture
def subject(
    state_view: StateView,
    equipment: EquipmentHandler,
) -> InitializeImpl:
    """Subject command implementation to test."""
    return InitializeImpl(state_view=state_view, equipment=equipment)


@pytest.mark.parametrize(
    "input_sample_wave_length, input_measure_mode", [([1, 2], "multi"), ([1], "single")]
)
async def test_absorbance_reader_implementation(
    decoy: Decoy,
    input_sample_wave_length: List[int],
    input_measure_mode: str,
    subject: InitializeImpl,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should validate, find hardware module if not virtualized, and disengage."""
    params = InitializeParams(
        moduleId="unverified-module-id",
        measureMode=input_measure_mode,  # type: ignore[arg-type]
        sampleWavelengths=input_sample_wave_length,
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)
    verified_module_id = AbsorbanceReaderId("module-id")

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when((absorbance_module_hw.supported_wavelengths)).then_return([1, 2])

    result = await subject.execute(params=params)

    decoy.verify(
        await absorbance_module_hw.set_sample_wavelength(
            ABSMeasurementMode(params.measureMode),
            params.sampleWavelengths,
            reference_wavelength=params.referenceWavelength,
        ),
        times=1,
    )
    assert result == SuccessData(
        public=InitializeResult(),
        state_update=update_types.StateUpdate(
            absorbance_reader_state_update=update_types.AbsorbanceReaderStateUpdate(
                module_id="module-id",
                initialize_absorbance_reader_update=update_types.AbsorbanceReaderInitializeUpdate(
                    measure_mode=input_measure_mode,  # type: ignore[arg-type]
                    sample_wave_lengths=input_sample_wave_length,
                    reference_wave_length=None,
                ),
            )
        ),
    )


@pytest.mark.parametrize(
    "input_sample_wave_length, input_measure_mode",
    [([1, 2, 3], "multi"), ([3], "single")],
)
async def test_initialize_raises_invalid_wave_length(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: InitializeImpl,
    input_sample_wave_length: List[int],
    input_measure_mode: str,
) -> None:
    """Should raise an InvalidWavelengthError error."""
    params = InitializeParams(
        moduleId="unverified-module-id",
        measureMode=input_measure_mode,  # type: ignore[arg-type]
        sampleWavelengths=input_sample_wave_length,
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    verified_module_id = AbsorbanceReaderId("module-id")
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when((absorbance_module_hw.supported_wavelengths)).then_return([1, 2])

    with pytest.raises(InvalidWavelengthError):
        await subject.execute(params=params)


@pytest.mark.parametrize(
    "input_sample_wave_length, input_measure_mode",
    [([], "multi"), ([], "single")],
)
async def test_initialize_raises_measure_mode_not_matching(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: InitializeImpl,
    input_sample_wave_length: List[int],
    input_measure_mode: str,
) -> None:
    """Should raise an error that the measure mode does not match sample wave."""
    params = InitializeParams(
        moduleId="unverified-module-id",
        measureMode=input_measure_mode,  # type: ignore[arg-type]
        sampleWavelengths=input_sample_wave_length,
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    verified_module_id = AbsorbanceReaderId("module-id")
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when((absorbance_module_hw.supported_wavelengths)).then_return([1, 2])

    with pytest.raises(ValueError):
        await subject.execute(params=params)


async def test_initialize_single_raises_reference_wave_length_not_matching(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: InitializeImpl,
) -> None:
    """Should raise an error that the measure mode does not match sample wave."""
    params = InitializeParams(
        moduleId="unverified-module-id",
        measureMode="single",
        sampleWavelengths=[1],
        referenceWavelength=3,
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    verified_module_id = AbsorbanceReaderId("module-id")
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when((absorbance_module_hw.supported_wavelengths)).then_return([1, 2])

    with pytest.raises(InvalidWavelengthError):
        await subject.execute(params=params)


async def test_initialize_multi_raises_no_reference_wave_length(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: InitializeImpl,
) -> None:
    """Should raise an error that the measure mode does not match sample wave."""
    params = InitializeParams(
        moduleId="unverified-module-id",
        measureMode="multi",
        sampleWavelengths=[1, 2],
        referenceWavelength=3,
    )

    mabsorbance_module_substate = decoy.mock(cls=AbsorbanceReaderSubState)
    verified_module_id = AbsorbanceReaderId("module-id")
    absorbance_module_hw = decoy.mock(cls=AbsorbanceReader)

    decoy.when(
        state_view.modules.get_absorbance_reader_substate("unverified-module-id")
    ).then_return(mabsorbance_module_substate)

    decoy.when(mabsorbance_module_substate.module_id).then_return(verified_module_id)

    decoy.when(equipment.get_module_hardware_api(verified_module_id)).then_return(
        absorbance_module_hw
    )

    decoy.when((absorbance_module_hw.supported_wavelengths)).then_return([1, 2])

    with pytest.raises(ValueError):
        await subject.execute(params=params)



"""Types that are common across protocols."""

from typing import TYPE_CHECKING, Type, TypeVar, Union

from opentrons.hardware_control.types import OT3Mount
from opentrons.types import Mount

if TYPE_CHECKING:
    from opentrons.config.types import OT3Config, RobotConfig


class OT2RobotType:
    pass


class FlexRobotType:
    pass


CalibrationType = TypeVar("CalibrationType")

MountArgType = TypeVar(
    "MountArgType", Mount, Union[OT3Mount, Mount], contravariant=True
)

# Use TYPE_CHECKING conditional to break circular import between
# opentrons.config.types and opentrons.hardware_control.protocols.types
if TYPE_CHECKING:
    ConfigType = TypeVar("ConfigType", "RobotConfig", "OT3Config")
else:
    ConfigType = TypeVar("ConfigType")

ProtocolRobotType = TypeVar(
    "ProtocolRobotType", Type[FlexRobotType], Type[OT2RobotType], covariant=True
)



"""Tests for opentrons.hardware_control.module_control."""

from typing import Awaitable, Callable, List, Union, cast

import pytest
from decoy import Decoy, matchers

from opentrons.drivers.rpi_drivers.interfaces import USBDriverInterface
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control import types
from opentrons.hardware_control.module_control import AttachedModulesControl
from opentrons.hardware_control.modules import AbstractModule
from opentrons.hardware_control.modules.types import (
    ModuleAtPort,
    ModuleType,
    SimulatingModuleAtPort,
)


@pytest.fixture()
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mocked out HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture()
def usb_bus(decoy: Decoy) -> USBDriverInterface:
    """Get a mocked out USBBus."""
    return decoy.mock(cls=USBDriverInterface)


@pytest.fixture()
def build_module(decoy: Decoy) -> Callable[..., Awaitable[AbstractModule]]:
    """Get a mocked out AttachedModuleControl.build_module.

    !!! warning

        This is bad testing practice and code smell. The fact that I can't
        test this module without faking out a part of it means that the
        `AttachedModulesControl` is doing too much work _and_ these tests
        are too brittle and of questionable value.
    """
    return cast(
        Callable[..., Awaitable[AbstractModule]],
        decoy.mock(name="build_module", is_async=True),
    )


@pytest.fixture()
def event_callback(decoy: Decoy) -> Callable[[types.HardwareEvent], None]:
    return decoy.mock(name="event_callback")  # type: ignore[no-any-return]


@pytest.fixture()
def subject(
    hardware_api: HardwareAPI,
    usb_bus: USBDriverInterface,
    build_module: Callable[..., Awaitable[AbstractModule]],
    event_callback: Callable[[types.HardwareEvent], None],
) -> AttachedModulesControl:
    modules_control = AttachedModulesControl(
        api=hardware_api, usb=usb_bus, event_callback=event_callback
    )

    # TODO(mc, 2022-03-01): partial patching the class under test creates
    # a contaminated test subject that reduces the value of these tests
    # https://github.com/testdouble/contributing-tests/wiki/Partial-Mock
    modules_control.build_module = build_module  # type: ignore[assignment]
    return modules_control


@pytest.mark.parametrize(
    "module_at_port_input",
    [
        ([ModuleAtPort(port="/dev/foo", name="bar")]),
        (
            [
                SimulatingModuleAtPort(
                    port="/dev/foo",
                    name="bar",
                    serial_number="test-123",
                    model="mymodel",
                )
            ]
        ),
    ],
)
async def test_register_modules(
    decoy: Decoy,
    usb_bus: USBDriverInterface,
    build_module: Callable[..., Awaitable[AbstractModule]],
    hardware_api: HardwareAPI,
    subject: AttachedModulesControl,
    module_at_port_input: Union[List[ModuleAtPort], List[SimulatingModuleAtPort]],
) -> None:
    """It should register attached modules."""
    actual_ports = [
        ModuleAtPort(
            port="/dev/foo",
            name="tempdeck",
            usb_port=USBPort(name="baz", port_number=0),
        )
    ]

    module = decoy.mock(cls=AbstractModule)
    decoy.when(module.usb_port).then_return(USBPort(name="baz", port_number=0))

    decoy.when(usb_bus.match_virtual_ports(module_at_port_input)).then_return(
        actual_ports
    )
    decoy.when(
        await build_module(
            port="/dev/foo",
            usb_port=USBPort(name="baz", port_number=0),
            type=ModuleType.TEMPERATURE,
            sim_serial_number=None,
            sim_model=None,
        )
    ).then_return(module)

    await subject.register_modules(new_mods_at_ports=module_at_port_input)
    result = subject.available_modules

    assert result == [module]


async def test_register_modules_sort(
    decoy: Decoy,
    usb_bus: USBDriverInterface,
    build_module: Callable[..., Awaitable[AbstractModule]],
    hardware_api: HardwareAPI,
    subject: AttachedModulesControl,
) -> None:
    """It should sort modules by port and hub, in ascending order."""
    module_1 = decoy.mock(cls=AbstractModule)
    decoy.when(module_1.usb_port).then_return(
        USBPort(name="a", port_number=4, hub=True, hub_port=2)
    )

    module_2 = decoy.mock(cls=AbstractModule)
    decoy.when(module_2.usb_port).then_return(
        USBPort(name="b", port_number=4, hub=True, hub_port=1)
    )

    module_3 = decoy.mock(cls=AbstractModule)
    decoy.when(module_3.usb_port).then_return(USBPort(name="c", port_number=3))

    module_4 = decoy.mock(cls=AbstractModule)
    decoy.when(module_4.usb_port).then_return(USBPort(name="x", port_number=2))

    module_5 = decoy.mock(cls=AbstractModule)
    decoy.when(module_5.usb_port).then_return(USBPort(name="z", port_number=1))

    new_mods_at_ports = [ModuleAtPort(port="/dev/foo", name="bar")]
    actual_ports = [
        ModuleAtPort(port="/dev/a", name="magdeck", usb_port=module_1.usb_port),
        ModuleAtPort(port="/dev/b", name="tempdeck", usb_port=module_2.usb_port),
        ModuleAtPort(port="/dev/c", name="thermocycler", usb_port=module_3.usb_port),
        ModuleAtPort(port="/dev/d", name="heatershaker", usb_port=module_4.usb_port),
        ModuleAtPort(
            port="/dev/d", name="absorbancereader", usb_port=module_5.usb_port
        ),
    ]

    decoy.when(usb_bus.match_virtual_ports(new_mods_at_ports)).then_return(actual_ports)

    for mod in [module_1, module_2, module_3, module_4, module_5]:
        decoy.when(
            await build_module(
                usb_port=mod.usb_port,
                port=matchers.Anything(),
                type=matchers.Anything(),
                sim_serial_number=None,
                sim_model=None,
            )
        ).then_return(mod)

    await subject.register_modules(new_mods_at_ports=new_mods_at_ports)
    result = subject.available_modules

    assert result == [module_5, module_4, module_3, module_2, module_1]



"""Test the loadLidStack command."""

import inspect

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.load_lid_stack import (
    LoadLidStackImplementation,
    LoadLidStackParams,
    LoadLidStackResult,
)
from opentrons.protocol_engine.execution import EquipmentHandler, LoadedLabwareData
from opentrons.protocol_engine.resources import labware_validation
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    LoadedLidStackUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    OnAddressableAreaLocationSequenceComponent,
    OnLabwareLocation,
    OnLabwareLocationSequenceComponent,
)
from opentrons.types import DeckSlotName


@pytest.fixture(autouse=True)
def patch_mock_labware_validation(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out labware_validations.py functions."""
    for name, func in inspect.getmembers(labware_validation, inspect.isfunction):
        monkeypatch.setattr(labware_validation, name, decoy.mock(func=func))


async def test_load_lid_stack(
    decoy: Decoy,
    tiprack_lid_def: LabwareDefinition,
    lid_stack_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """It should load a lid onto a labware."""
    subject = LoadLidStackImplementation(equipment=equipment, state_view=state_view)
    input_location = DeckSlotLocation(slotName=DeckSlotName.SLOT_A1)

    data = LoadLidStackParams(
        location=input_location,
        loadName="someLid",
        namespace="someNamespace",
        version=1,
        quantity=3,
    )

    decoy.when(
        state_view.addressable_areas.raise_if_area_not_in_deck_configuration("A1")
    ).then_return(True)
    decoy.when(
        state_view.geometry.ensure_location_not_occupied(location=input_location)
    ).then_return(input_location)

    decoy.when(
        await equipment.load_labware(
            load_name="protocol_engine_lid_stack_object",
            namespace="opentrons",
            version=1,
            location=input_location,
            labware_id=None,
        )
    ).then_return(
        LoadedLabwareData(
            labware_id="stack-labware-id",
            definition=lid_stack_def,
            offsetId=None,
        )
    )
    decoy.when(
        labware_validation.validate_definition_is_system(lid_stack_def)
    ).then_return(True)

    lids = [
        LoadedLabwareData(
            labware_id="lid-labware-1", definition=tiprack_lid_def, offsetId=None
        ),
        LoadedLabwareData(
            labware_id="lid-labware-2",
            definition=tiprack_lid_def,
            offsetId=None,
        ),
        LoadedLabwareData(
            labware_id="lid-labware-3", definition=tiprack_lid_def, offsetId=None
        ),
    ]

    decoy.when(
        await equipment.load_lids(
            load_name="someLid",
            namespace="someNamespace",
            version=1,
            location=OnLabwareLocation(labwareId="stack-labware-id"),
            quantity=3,
            labware_ids=None,
        )
    ).then_return(lids)

    stack_loc = OnAddressableAreaLocationSequenceComponent(addressableAreaName="A1")
    on_stack_loc = OnLabwareLocationSequenceComponent(
        labwareId="stack-labware-id", lidId=None
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(input_location)
    ).then_return([stack_loc])

    result = await subject.execute(data)
    expected_result = SuccessData(
        public=LoadLidStackResult(
            stackLabwareId="stack-labware-id",
            labwareIds=["lid-labware-1", "lid-labware-2", "lid-labware-3"],
            definition=tiprack_lid_def,
            lidStackDefinition=lid_stack_def,
            location=input_location,
            stackLocationSequence=[stack_loc],
            locationSequences=[
                [on_stack_loc, stack_loc],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="lid-labware-1", lidId=None
                    ),
                    on_stack_loc,
                    stack_loc,
                ],
                [
                    OnLabwareLocationSequenceComponent(
                        labwareId="lid-labware-2", lidId=None
                    ),
                    OnLabwareLocationSequenceComponent(
                        labwareId="lid-labware-1", lidId=None
                    ),
                    on_stack_loc,
                    stack_loc,
                ],
            ],
        ),
        state_update=StateUpdate(
            loaded_lid_stack=LoadedLidStackUpdate(
                stack_id="stack-labware-id",
                stack_object_definition=lid_stack_def,
                stack_location=input_location,
                definition=tiprack_lid_def,
                new_locations_by_id={
                    "lid-labware-1": OnLabwareLocation(labwareId="stack-labware-id"),
                    "lid-labware-2": OnLabwareLocation(labwareId="lid-labware-1"),
                    "lid-labware-3": OnLabwareLocation(labwareId="lid-labware-2"),
                },
            )
        ),
    )
    assert result.public.locationSequences == expected_result.public.locationSequences
    assert result == result



"""Test Thermocycler close lid command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_engine.commands import thermocycler as tc_commands
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.thermocycler.close_lid import (
    CloseLidImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import MotorAxis


async def test_close_lid(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    movement: MovementHandler,
) -> None:
    """It should be able to close the specified module's lid."""
    subject = CloseLidImpl(
        state_view=state_view, equipment=equipment, movement=movement
    )

    data = tc_commands.CloseLidParams(moduleId="input-thermocycler-id")
    expected_module_id = ThermocyclerModuleId("thermocycler-id")
    expected_result = tc_commands.CloseLidResult()

    tc_module_substate = decoy.mock(cls=ThermocyclerModuleSubState)
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(
        state_view.modules.get_thermocycler_module_substate("input-thermocycler-id")
    ).then_return(tc_module_substate)

    decoy.when(tc_module_substate.module_id).then_return(expected_module_id)

    # Get attached hardware modules
    decoy.when(equipment.get_module_hardware_api(expected_module_id)).then_return(
        tc_hardware
    )
    decoy.when(state_view.motion.get_robot_mount_axes()).then_return(
        [MotorAxis.EXTENSION_Z]
    )
    result = await subject.execute(data)

    decoy.verify(
        await movement.home([MotorAxis.X, MotorAxis.Y, MotorAxis.EXTENSION_Z]),
        await tc_hardware.close(),
        times=1,
    )
    assert result == SuccessData(
        public=expected_result,
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Test get tip presence commands."""

import pytest
from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.get_tip_presence import (
    GetTipPresenceImplementation,
    GetTipPresenceParams,
    GetTipPresenceResult,
)
from opentrons.protocol_engine.execution import TipHandler
from opentrons.protocol_engine.types import TipPresenceStatus


@pytest.mark.parametrize(
    "status",
    [
        TipPresenceStatus.PRESENT,
        TipPresenceStatus.ABSENT,
        TipPresenceStatus.UNKNOWN,
    ],
)
async def test_get_tip_presence_implementation(
    decoy: Decoy,
    tip_handler: TipHandler,
    status: TipPresenceStatus,
) -> None:
    """A GetTipPresence command should have an execution implementation."""
    subject = GetTipPresenceImplementation(tip_handler=tip_handler)
    data = GetTipPresenceParams(
        pipetteId="pipette-id",
    )

    decoy.when(
        await tip_handler.get_tip_presence(
            pipette_id="pipette-id",
        )
    ).then_return(status)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=GetTipPresenceResult(status=status),
    )



"""Test pause command."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.wait_for_duration import (
    WaitForDurationImplementation,
    WaitForDurationParams,
    WaitForDurationResult,
)
from opentrons.protocol_engine.execution import RunControlHandler


async def test_pause_implementation(
    decoy: Decoy,
    run_control: RunControlHandler,
) -> None:
    """It should await the duration using the RunControlHandler."""
    subject = WaitForDurationImplementation(run_control=run_control)

    data = WaitForDurationParams(message="hello world", seconds=42.0)

    result = await subject.execute(data)

    assert result == SuccessData(public=WaitForDurationResult())
    decoy.verify(await run_control.wait_for_duration(42.0), times=1)



from datetime import datetime

from ..tasks import AbstractTaskCore


class LegacyTaskCore(AbstractTaskCore):
    def __init__(self) -> None:
        pass

    def get_created_at_timestamp(self) -> datetime:
        raise NotImplementedError("Legacy protocols do not implement tasks.")

    def is_done(self) -> bool:
        raise NotImplementedError("Legacy protocols do not implement tasks.")

    def is_started(self) -> bool:
        raise NotImplementedError("Legacy protocols do not implement tasks.")

    def get_finished_at_timestamp(self) -> datetime | None:
        raise NotImplementedError("Legacy protocols do not implement tasks.")



"""Test setTipState command."""

import pytest
from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.set_tip_state import (
    SetTipStateImplementation,
    SetTipStateParams,
    SetTipStateResult,
)
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import TipRackWellState


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


async def test_set_tip_state(decoy: Decoy, mock_state_view: StateView) -> None:
    """A setTipState should have an execution implementation."""
    subject = SetTipStateImplementation(state_view=mock_state_view)

    params = SetTipStateParams(
        labwareId="labware-id",
        wellNames=["well-1", "well-2"],
        tipWellState=TipRackWellState.USED,
    )

    result = await subject.execute(params)

    decoy.verify(
        mock_state_view.labware.raise_if_not_tip_rack("labware-id"),
        mock_state_view.labware.raise_if_wells_are_invalid(
            labware_id="labware-id",
            well_names=["well-1", "well-2"],
        ),
    )

    assert result == SuccessData(
        public=SetTipStateResult(),
        state_update=update_types.StateUpdate(
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.USED,
                labware_id="labware-id",
                well_names=["well-1", "well-2"],
            )
        ),
    )
